/* 
 * Program:             $RCSfile: xbd2.cc,v $  $Revision: 41.3 $
 * 
 * Purpose:             XML builder.
 *
 * Author:              H.Yoshida  2002/06/16
 *                      S.Yamamoto  2002/08/04
 *
 * (C) Copyright:       H.Yoshida and S.Yamamoto  2002 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: xbd2.cc,v 41.3 2003/05/29 04:20:43 yamamoto Exp yamamoto $";
#endif

extern "C" {
#include <stdio.h>
#include <Sapid/Sapid.h>
}

#include <algorithm>
#include <list>
#include <string>
#include <vector>

// #define DEBUG_INPUT

using namespace std;

// ---------------------------------------------------------------------

// Node.

#define BIT_WIDTH_WEIGHT    8 

struct XMLNODE {
    string      name;
    string      attrs;
    long        length;     // (bitWidthOf(long) - BIT_WIDTH_WEIGHT) bit.
    short       weight;     // BIT_WIDTH_WEIGHT (define as 8) bit.
    
    XMLNODE() : length(0), weight(0) {}
};

typedef list<XMLNODE> XMLNODES;

// Tag type.

enum TAGSORT {
    T_END, T_EMPTY, T_START
};

// Insert position.

struct INSPTR {
    long        offset;
    TAGSORT     type;
    XMLNODE     *lpnode;
};

typedef vector<INSPTR> INSPTRS;

// ---------------------------------------------------------------------

// File names.
static char     _src_path[FILENAME_MAX + 1];
static char     _meta_path[FILENAME_MAX + 1];
static char     dst_file[FILENAME_MAX + 1];

// Line counter.
static int      _linecount = 0;

// Option flags.
static bool     _opt_no_textnode = false;
static bool     _opt_entityref1 = false;
static bool     _opt_entityref2 = false;
static bool     _is_out_to_file = false;
static bool     _is_verbose = false;
static bool     _is_Verbose = false;

// Text node name.
static char     _textnode_name[] = "T";

// This program name.
static SpdConstString   program_name;

static void print_usage(SpdConstString command_name);

// ---------------------------------------------------------------------

// Desc: ֤դؿ

bool ins_ptr_sorter(const INSPTR &r1, const INSPTR &r2)
{
    if (r1.offset == r2.offset) {
        if (r1.type == r2.type) {
            long    &l1 = r1.lpnode->length;
            long    &l2 = r2.lpnode->length;
            short   &w1 = r1.lpnode->weight;
            short   &w2 = r2.lpnode->weight;

            if (r1.type == T_START) {
                // if (l1 == l2) return (w1 > w2);
                // return (l1 > l2);
                return ((l1 == l2) ? (w1 > w2) : (l1 > l2));
            } else if (r1.type == T_END) {
                // if (l1 == l2) return (w1 < w2);
                // return (l1 < l2);
                return ((l1 == l2) ? (w1 < w2) : (l1 < l2));
            } else if (r1.type == T_EMPTY) {
                return (w1 < w2);
            }
            assert(false);
        }
        return (r1.type < r2.type);
    }
    return (r1.offset < r2.offset);
}

// Desc: element tableɤƳǼɲä롥

bool load_metainfo_file(FILE *fp, FILE *fpout,
                        XMLNODES &node_list, INSPTRS &ins_ptr_list)
{
    int     line = 0;

    if (_is_verbose) {
        spdMessage("Loading element table.\n");
    }

    while (true) {
        const int   BSIZE = FILENAME_MAX + 1;
        char        buf[BSIZE];

        // Read a line.
        fgets(buf, BSIZE, fp);
        if (feof(fp)) {
            break;
        }
        line++;

        if (*buf == '#') {
            // Skip comment.
            continue;
        } else if (*buf == '!') {
            // XML string.
            fputs(buf + 1, fpout);
            continue;
        }

        // Scan a line.
        char    name[BSIZE], attrs[BSIZE];
        int     offset, length, weight, ap, col;

        col = sscanf(buf, "%ld %ld %d %s %n",
                     &offset, &length, &weight, name, &ap);
        if (_is_Verbose) {
            // spdMessage("Scan line %d: %s", line, buf);
            spdMessage("line %4d: %4d, %4d, %2d, \"%s\", %d.\n",
                       line, offset, length, weight, name, ap);
        }

        if (col < 4 || length < 0) {
            spdWarning("%d, %d, %d, \"%s\", %d: Invalid format in l.%d.\n",
                       offset, length, weight, name, ap, line);
            continue;
        }

        // Add node.
        XMLNODE     node;

        node.name = string(name);
        node.length = length;
        node.weight = weight;

        if (ap < strlen(buf)) {
            char    *l = buf + strlen(buf) - 1;
            
            if (*(l - 1) == '\r') {
                *(l - 1) = '\0';
            } else if (*l == '\n') {
                *l = '\0';
            }
            strcpy(attrs, buf + ap);
            node.attrs = " " + string(attrs);
        }
        
        node_list.push_back(node);

        // Add insert position.
        INSPTR  ipr;

        ipr.lpnode = &node_list.back();
        ipr.offset = offset;
        if (0 < length) {
            ipr.type = T_START;
            ins_ptr_list.push_back(ipr);    // Start tag.
            ipr.offset = offset + length;
            ipr.type = T_END;
            ins_ptr_list.push_back(ipr);    // End tag.
        } else {
            ipr.type = T_EMPTY;
            ins_ptr_list.push_back(ipr);    // Empty tag.
        }
    }

    if (node_list.size() == 0) {
        spdError("No elements in element table.\n");
        return false;
    }

    if (_is_verbose) {
        spdMessage("Number of nodes in input file = %d.\n", node_list.size());
    }

    // Sort tag position.
    sort(ins_ptr_list.begin(), ins_ptr_list.end(), ins_ptr_sorter);

#if defined(DEBUG_INPUT)
    int                     i = 0;
    INSPTRS::const_iterator ip;
    for (ip = ins_ptr_list.begin(); ip < ins_ptr_list.end(); ip++) {
        spdMessage("%4d: offset %4d, weight %2d, type %d, name \"%s\".\n",
                   i, ip->offset, ip->lpnode->weight, ip->type,
                   ip->lpnode->name.c_str());
        i++;
    }
#endif
    
    // Check root node.
    if (ins_ptr_list.begin()->offset != 0) {
        spdErrorExit("The first tag position isn't zero.\n");
    }

    return true;
}

// Desc: ƥȥΡɤ򸡽Фɲä롥

void merge_text_node(XMLNODES &node_list, INSPTRS &ins_ptr_list)
{
    if (_is_verbose) {
        spdMessage("Merging text node.\n");
    }
    assert(1 < ins_ptr_list.size());

    INSPTRS ins_ptr_list_old;
    INSPTRS ins_ptr_list_text;

    // create text node buffer
    INSPTRS::const_iterator ip;
    for (ip = ins_ptr_list.begin(); ip < ins_ptr_list.end() - 1; ip++) {
        INSPTRS::const_iterator ip2 = ip + 1;
        const long  &off1 = ip->offset;
        const long  &off2 = ip2->offset;

        assert(off1 <= off2);

        if ((off1 < off2) && (ip->type != T_START || ip2->type != T_END)) {
            // Add text node.
            XMLNODE     node;

            node.name = string(_textnode_name);
            node.length = off2 - off1;
            node_list.push_back(node);

            // Add insert position.
            INSPTR  ipr;

            ipr.lpnode = &node_list.back();
            ipr.offset = off1;
            ipr.type = T_START;
            ins_ptr_list_text.push_back(ipr);   // Start tag.
            ipr.offset = off2;
            ipr.type = T_END;
            ins_ptr_list_text.push_back(ipr);   // End tag.
        }
    }

    if (_is_verbose) {
        spdMessage("Number of nodes after text node merging = %d.\n",
                   node_list.size());
    }

    // Merge text tag position.
    if (0 < ins_ptr_list_text.size()) {
        // Copy original.
        INSPTRS::size_type s = ins_ptr_list.size();
        ins_ptr_list_old.resize(s);
        copy(ins_ptr_list.begin(), ins_ptr_list.end(),
             ins_ptr_list_old.begin());

        // Expand buffer area.
        INSPTRS::size_type sn = s + ins_ptr_list_text.size();
        ins_ptr_list.clear();
        ins_ptr_list.resize(sn);

        // Merge original and text nodes.
        merge(ins_ptr_list_old.begin(), ins_ptr_list_old.end(),
              ins_ptr_list_text.begin(), ins_ptr_list_text.end(),
              ins_ptr_list.begin(), ins_ptr_sorter);
    }

    return;
}

// Desc: XMLʸʸϤ롥

void output_xml_char(FILE *fpout, const int ch)
{
    static bool     crpred = false; // 'CR'

    if (!crpred && (ch == '\n' || ch == '\r')) {
        _linecount++;
    }
    crpred = (ch == '\r');

    char    buf[16];

    buf[0] = ch;
    buf[1] = '\0';
    switch (ch) {               // üʸ
      case '<':
        strcpy(buf, "&lt;");
        break;
      case '>':
        strcpy(buf, "&gt;");
        break;
      case '&':
        strcpy(buf, "&amp;"); 
        break;
      case 0x9:                 // Horizontal tab.
        if (_opt_entityref1) strcpy(buf, "&#x9;");
        break;
      case 0x20:                // Space.
        if (_opt_entityref1) strcpy(buf, "&#x20;");
        break;
      case '\n':
        if (_opt_entityref2) strcpy(buf, "&#xA;");
        break;
      case '\r':
        if (_opt_entityref2) strcpy(buf, "&#xD;");
        break;
    }
    fputs(buf, fpout);
}

// Desc: XMLʸ˥Ϥ롥

void output_xml_tag(FILE *fpout, const XMLNODE *lpnode, TAGSORT type)
{
    const char  *tn = lpnode->name.c_str();
    const char  *ta = lpnode->attrs.c_str();

    switch (type) {
      case T_START:
        fprintf(fpout, "<%s%s>", tn, ta);
        break;
      case T_END:
        fprintf(fpout, "</%s>", tn);
        break;
      case T_EMPTY:
        fprintf(fpout, "<%s%s/>", tn, ta);
        break;
    }
}

// Desc: ѴΥᥤؿ

bool process_main(FILE *fp, FILE *fpout, const INSPTRS &ins_ptr_list)
{
    int     pos = 0;
    int     ch;

    INSPTRS::const_iterator ip;
    for (ip = ins_ptr_list.begin(); ip < ins_ptr_list.end(); ip++) {
        const long &nextpos = ip->offset;

        // Copy characters until next inset position.
        for (int i = 0; i < nextpos - pos; i++) {
            ch = getc(fp);
            if (ch == EOF) {
                goto err_eof;
            }
            output_xml_char(fpout, ch);
        }
        pos = nextpos;
        
        // Insert tag.
        output_xml_tag(fpout, ip->lpnode, ip->type);
    }

    // It must not read next character.
    if (getc(fp) != EOF) {
        spdWarning("The last tag position isn't the end of file.\n");
    }

    if (_is_verbose) {
        // spdMessage("Markup succeeded.\n");
        spdMessage("Line counter = %d.\n", _linecount);
    }

    return true;

    /* Error occurs. */ 
  err_eof:
    spdError("Detect the end of file before the last tag position.\n");
    return false;
}

// Desc: ȥݥȡ

bool parse_param(int argc, char *argv[])
{
    int     ch;

    while((ch = getopt(argc, argv, "d:nr:vV")) != -1) {
        switch (ch) {
          case 'd':
            _is_out_to_file = true;
            strcpy(dst_file, optarg);
            break;
          case 'n':
            _opt_no_textnode = true;
            break;
          case 'r':
            if (SPD_STREQ(optarg, "1")) {
                _opt_entityref1 = true;
            } else if (SPD_STREQ(optarg, "2")) {
                _opt_entityref2 = true;
            } else {
                spdError("Invalid \"-r\" option.\n");
                print_usage(program_name);
            }
            break;
          case 'v':
            _is_verbose = true;
            break;
          case 'V':
            _is_Verbose = true;
            break;
          default:
            spdError("'%c': Invalid option.\n", ch);
            print_usage(program_name);
            break;
        }
    }
    argc -= optind;
    argv += optind;
    if (argc != 2) {
        spdError("No arguments.\n");
        print_usage(program_name);
    }
    strcpy(_src_path, argv[0]);
    strcpy(_meta_path, argv[1]);
    return true;
}

// Desc: ȥݥȡ

int main(int argc, char *argv[])
{
    bool    retb;
    FILE    *fpsrc;
    FILE    *fpmeta;
    FILE    *fpout;

    program_name = spdBasename(argv[0]);
    spdInitialize(program_name);

    if (!parse_param(argc, argv)) {
        print_usage(program_name);
    }

    // Open source file.
    fpsrc = spdFopen(_src_path, SPD_FOPEN_RB_MODE);

    // Open element table.
    fpmeta = spdFopen(_meta_path, SPD_FOPEN_RT_MODE);

    // Open output file.
    if (_is_out_to_file) {
        strcat(dst_file, "/");
        strcat(dst_file, spdBasename(spdStrdup(_src_path)));
        strcat(dst_file, ".xml");
        fpout = spdFopen(dst_file, SPD_FOPEN_WB_MODE);
    } else {
        fpout = stdout;
    }

    // Ρɥꥹȡ
    XMLNODES node_list;

    // ֡
    INSPTRS ins_ptr_list;

    // Load element table.
    retb = load_metainfo_file(fpmeta, fpout, node_list, ins_ptr_list);
    if (!retb) {
        spdErrorExit("load_metainfo_file(): %d: Failed.\n", retb);
    }

    if (!_opt_no_textnode) {
        merge_text_node(node_list, ins_ptr_list);
    }

    // Build process.
    retb = process_main(fpsrc, fpout, ins_ptr_list);
    if (!retb) {
        spdErrorExit("process_main(): %d: Failed.\n", retb);
    }

    // Close files.
    spdFclose(fpsrc);
    spdFclose(fpmeta);
    spdFclose(fpout);

    return (EXIT_SUCCESS);
}

static void print_usage(SpdConstString command_name)
{
    const char *usageMessage =
        "\n"
        "NAME\n"
        "    %s\n"
        "SYNOPSIS\n"
        "    %s [-fn][-r1][-r2] <targetJavaFile> <elementTable>\n"
        "DESCRIPTION\n"
        "    \n"
        "NOTE\n"
        "    elementTable format is:\n"
        "    offset:%%d length:%%d weight:%%d name:%%s [attrs:%%s]\n"
        "VERSION\n"
        "    %d.%d.%d\n";

    spdErrorExit(usageMessage,
                 command_name, command_name,
                 SapidVersion, SapidRevision, SapidPatchLevel);
}
