/*
 * Created on 2003/06/07
 * SpdElement.java - Copyright (c) 2003 H.Yoshida
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS IS FREEWARE. NO WARRANTY.
 */

package JXModel3.xsdml;

import org.w3c.dom.*;

/**
 * XSDML Element Class
 * @author H.Yoshida
 */
public class SpdElement 
{
	static final String ELMNL = "nl";
	static final String ATTLINE = "line";
	
	/**
	 * The attached XML element object.
	 * @see http://java.sun.com/j2se/1.4/ja/docs/ja/api/org/w3c/dom/Element.html
	 */
	public Element domelm = null;

	/**
	 * The constructor.
	 * Attach a XML element.
	 */
	public SpdElement(Element elm) 
	{
		domelm = elm;
	}

	/**
	 * Get element name.
	 * @return
	 */
	public String getName() 
	{
		return domelm.getTagName();
	}

	/**
	 * Get 'id' attribute value.
	 * @return
	 */
	public String getID() 
	{
		return domelm.getAttribute("id");
	}

	/**
	 * Get 'sort' attribute value.
	 * @return
	 */
	public String getSort() 
	{
		return domelm.getAttribute("sort");
	}

	/**
	 * Get text representation.
	 * @return
	 */
	public String getText() 
	{
		String ret = "";
		NodeList nl = domelm.getChildNodes();
		for (int i=0; i<nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType()==Node.ELEMENT_NODE) {
				Element e = (Element)n;
				ret += new SpdElement(e).getText();
			}
			if (n.getNodeType()==Node.TEXT_NODE) {
				ret += n.getNodeValue();
			}
		}
		return ret;
	}

	/**
	 * Get child elements.
	 */
	public SpdElementList getChildren() 
	{
		SpdElementList el = new SpdElementList();

		NodeList nl = domelm.getChildNodes();
		for (int i=0; i<nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType()==Node.ELEMENT_NODE) {
				SpdElement e = new SpdElement((Element)n);
				el.add(e);
			}
		}
		
		return el;
	}

	/**
	 * Get child element.
	 */
	public SpdElement getChild(int idx)
	{
		if (idx<0) return null;
		
		SpdElementList el = getChildren();
		if (el.count()<=idx) return null;
		
		return el.getElementAt(idx);
	}

	/**
	 * Get parent element.
	 */
	public SpdElement getParent() 
	{
		Node n = domelm.getParentNode();
		if (n==null) return null;

		if (n.getNodeType()==Node.ELEMENT_NODE) {
			SpdElement e = new SpdElement((Element)n);
			return e;
		}
		return null;
	}

	/**
	 * Get next sibling element.
	 */
	public SpdElement getNextSibling() 
	{
		Node n = domelm.getNextSibling();
		
		while (n!=null) {
			if (n.getNodeType()==Node.ELEMENT_NODE) {
				SpdElement e = new SpdElement((Element)n);
				return e;
			}
			n = n.getNextSibling();
		}
		return null;
	}

	/**
	 * Get previous sibling element.
	 */
	public SpdElement getPreviousSibling() 
	{
		Node n = domelm.getPreviousSibling();
		
		while (n!=null) {
			if (n.getNodeType()==Node.ELEMENT_NODE) {
				SpdElement e = new SpdElement((Element)n);
				return e;
			}
			n = n.getPreviousSibling();
		}
		return null;
	}

	/**
	 * Get decendant elements.
	 */
	public SpdElementList getDecendant() 
	{
		SpdElementList el = new SpdElementList();

		getDecendantR(el); // recursive
		return el;
	}
	private SpdElementList getDecendantR(SpdElementList el) 
	{
		NodeList nl = domelm.getChildNodes();
		for (int i=0; i<nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType()==Node.ELEMENT_NODE) {
				SpdElement e = new SpdElement((Element)n);
				el.add(e);
				e.getDecendantR(el); // recursive
			}
		}
		
		return el;
	}

	/**
	 * Get preceding element.
	 */
	public SpdElement getPreceding() 
	{
		SpdElement e;
		
		e = getPreviousSibling();
		if (e==null) {
			return getParent();
		} else {
			while (true) {
				SpdElement ce = e.getChildren().last();
				if (ce==null) return e;
				e = ce;
			}
		}
	}

	/**
	 * Get following element.
	 */
	public SpdElement getFollowing()
	{
		SpdElement e, ce;
		
		ce = getChildren().first();
		if (ce==null) {
			e = this;
			while (true) {
				ce = e.getNextSibling();
				if (ce==null) {
					e = e.getParent();
					if (e==null) return null;
				} else {
					return ce;
				}
			}
		} else {
			return ce;
		}
	}

	/**
	 * Get line counter.
	 * @return
	 */
	public int getLine() 
	{
		// NEWLINE
		if (getName().equals(ELMNL)) {
			String linestr = domelm.getAttribute(ATTLINE);
			int line = Integer.valueOf(linestr).intValue();
			return line;
		}
		
		// NOT NEWLINE
		SpdElement elm = this;
		while (true) {
			elm = elm.getPreceding();
			if (elm==null) return 1; // TOP LINE
			if (elm.getName().equals(ELMNL)) {
				String linestr = elm.domelm.getAttribute(ATTLINE);
				int pastline = Integer.valueOf(linestr).intValue();
				return pastline + 1;
			}
		}
	}
}


