/* 
 * Program:             $RCSfile: FrejaEntry.java,v $  $Revision: 41.39 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrejaEntry.java,v 41.39 2003/03/27 09:05:43 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

/**
 * FrejaEntry is a entry pointer of freja (SDB).
 * A freja must be accessed through FrejaEntry.
 * You may recall that in C language, an file is accessed
 * through file pointer (FILE*).
 * FrejaEntry also treats with a source program.
 * You can get string of source program through FrejaEntry.
 */

public class FrejaEntry {
    private final static int    MAX_FREJA = FrjID.MAX_FREJA;
    private static boolean      freeFrejaID[];

    private int         id;
    private File        frejaDir;
    private String      frejaPath = null;
    private String      sourceFileName = null;
    private String      className = null;
    private String      canonicalName = null;
    private String      packageName = null;
    private String      encoding = null;
    private int         fileSize = -1;  /* Invalid size. */

    private final static int CRNum = FrjWorld.dbTable.getDBClassRelNum();
    private int         maxCRID[];

    static boolean debug = false;

    // Contents of this file.
    private ModifiableStringBuffer      text = null;
    private Vector                      changes = new Vector();
    private boolean                     modified = false;
    private JpdInfo                     jpdInfo;

    public JpdSDBformat                 jpdSDBformat;

    // Static methods for Freja Id.
    private static int getFreeID() {
        if (FrjWorld.isFidToZero()) {
            return 0;
        }

        // 0 is reserved for new objects.
        for (int i = 1; i < MAX_FREJA; i++) {
            if (freeFrejaID[i]) {
                freeFrejaID[i] = false;
                return i;
            }
        }

        FrjWorld.errorExit("Too many Frejas.");
        return -1;
    }

    static void freeFrejaID(FrejaEntry entry) {
        FrjWorld.debug(debug, "Free id: " + entry.getFrejaID() + " " +
                       entry.getCanonicalFrejaCode());
        freeFrejaID[entry.getFrejaID()] = true;
    }

    static void freeAllFrejaID() {
        // 0 is reserved for new objects.
        for (int i = 1; i < MAX_FREJA; i++) {
            freeFrejaID[i] = true;
        }
    }

    static {
        freeFrejaID = new boolean[MAX_FREJA];
        // 0 is reserved for new objects.
        for (int i = 1; i < MAX_FREJA; i++) {
            freeFrejaID[i] = true;
        }
    }

    // Constructors.
    /**
     * Create an access pointer of freja of <I>className</I>.
     * @param frejaPath  a directory path of freja 
     * @param className  a class name which you want to read
     *  A className is not dot-trailed style(e.g. java.util.Vector),
     *  but single identifier style(e.g. Vector).
     *  If you want to deal with dot-trailed style, please use Creja.
     */

    public FrejaEntry(String frejaPath, String className) {
        this(frejaPath, className, null);
    }

    public FrejaEntry(String frejaPath, String className, String pkgName) {
        id = getFreeID();
        this.frejaPath = frejaPath;
        this.packageName = pkgName;
        if (className.endsWith(".java")) {
            int index = className.lastIndexOf(".java");
            this.className = className.substring(0, index);
        } else {
            this.className = className;
        }

        frejaDir = new File(getFrejaPath());

        maxCRID = new int[CRNum];
        for (int i = 0; i < CRNum; i++) {
            maxCRID[i] = 0;
        }

        /* Added by K.Hoshino 2001/10/08. */
        if (exists()) { 
            jpdSDBformat = new JpdSDBformat();
            readJpdInfo();
        }
        setSourceFileInfo();
        canonicalName = frejaPath + File.separator +
                        (sourceFileName != null ? sourceFileName : className);

        // !! I don't handle RandomAccessFile well.
        // try {
        //     sourceFile = new RandomAccessFile(this.fileName, "r");
        // }
        // catch (IOException e) {
        //     FrjWorld.warning("Can't read " + this.fileName + ".");
        // }
        // catch (Exception e) {
        //     FrjWorld.warning("Error occurred when reading file");
        // }

        FrjWorld.debug(debug, "Get id: " + id + " " + getCanonicalFrejaCode());
    }

    /**
     * Open and read InfoFile on Freja directory and
     * set format for reading Freja.
     *
     * See "lib/SMU/getAProgramName.c".
     */

    private void readJpdInfo() {
        jpdInfo = new JpdInfo((FrejaEntry)this);

        // Set format.
        String sdbFormatInfoVal = jpdInfo.jpdGetInfoString("BINARY_FORMAT_SDB");
        if (sdbFormatInfoVal.equals("TRUE")) {
            jpdSDBformat.jpdSetReadSDBformatBinary();
            jpdSDBformat.jpdSetWriteSDBformatBinary();
        } else {
            jpdSDBformat.jpdSetReadSDBformatText();
            jpdSDBformat.jpdSetWriteSDBformatText();
        }

        // Check SPD_FIRST_CID_J_MODEL.
        int cid = jpdInfo.jpdGetInfoInt("SPD_FIRST_CID_J_MODEL");
        if (cid != SymSort.SPD_FIRST_CID_J_MODEL) {
            FrjWorld.errorExit(cid + ": SPD_FIRST_CID_J_MODEL should be " +
                               SymSort.SPD_FIRST_CID_J_MODEL + ".");
        }

        // Check SPD_[CR]ID_NUM_J_MODEL.
        int cidn = jpdInfo.jpdGetInfoInt("SPD_CID_NUM_J_MODEL");
        int ridn = jpdInfo.jpdGetInfoInt("SPD_RID_NUM_J_MODEL");
        if (cidn != FrjWorld.dbTable.getDBClassNum()) {
            FrjWorld.errorExit(cidn + ": SPD_CID_NUM_J_MODEL should be " +
                               FrjWorld.dbTable.getDBClassNum() + ".");
        }
        if (ridn != FrjWorld.dbTable.getDBRelNum()) {
            FrjWorld.errorExit(ridn + ": SPD_RID_NUM_J_MODEL should be " +
                               FrjWorld.dbTable.getDBRelNum() + ".");
        }

        // Check SapidSDBVersion.
        // See "Sapid/version.h" and "Japid3/Japid/Freja/Version.java".
        int sdb_version = jpdInfo.jpdGetInfoInt("SapidSDBVersion");
        if (sdb_version != Version.SapidSDBVersion) {
            FrjWorld.warning(sdb_version + ": SapidSDBVersion should be " + 
                             Version.SapidSDBVersion + ".");
        }
    }

    /**
     * Reset and save InfoFile on Freja directory.
     */
    public void saveJpdInfo() {
        // Save objects.
        for (Enumeration e = FrjExtents.getClassNames(); e.hasMoreElements();) {
            String name = (String)e.nextElement();
            int counter = 0;
            for (Enumeration ee = FrjExtents.getObjects(this, name);
                 ee.hasMoreElements(); /* Empty. */) {
                ee.nextElement();
                counter++;
            }
            jpdInfo.jpdSetInfoInt(name, counter);
        }
        // Save relations.
        FrjRelation saver = new FrjRelation();
        for (Enumeration e = FrjExtents.getRelationNames();
             e.hasMoreElements(); /* Empty. */) {
            String relName = (String)e.nextElement();
            int counter = 0;
            for (Enumeration ee = FrjExtents.getRelations(this, relName);
                 ee.hasMoreElements(); /* Empty. */) {
                ee.nextElement();
                counter++;
            }
            jpdInfo.jpdSetInfoInt(relName, counter);
        }
        jpdInfo.writeAllInfo();
    }

    /** Returns true if the Freja exists and you can read it. */
    public boolean exists() {
        if (frejaDir.exists() && frejaDir.isDirectory() && frejaDir.canRead()) {
            File file = new File(frejaDir, FrjFile.getClassNameStatically());
            if (file.exists() && file.isFile() && file.canRead()) {
                return true;
            }
        }
        return false;
    }

    /** Returns Id of this entry. */
    public int getFrejaID() {
        return id;
    }

    /** Returns directory path of Freja. */
    public String getFrejaPath() {
        return frejaPath + File.separator + className;
    }

    /** Returns canonical name of Freja. */
    String getCanonicalFrejaCode() {
        return canonicalName;
    }

    /**
     * Checks equality of frejas.
     */
    public boolean equals(FrejaEntry entry) {
        return getCanonicalFrejaCode().equals(entry.getCanonicalFrejaCode());
    }

    /*
     * readFile() doesn't work well.
     * JCOM return file's contents specified offset and length.
     */
    // public String readFile(int offset, int length) {
    //     byte    buff[] = new byte[128 * 4096];
    //     int     readByte;
    // 
    //     try {
    //         readByte = sourceFile.read(buff, offset, length);
    //         if (readByte < 0) {
    //             FrjWorld.warning("Error occurred when reading file.");
    //             return null;
    //         }
    //         return new String(buff, 0, 0, readByte);
    //     } catch (IOException e) {
    //         FrjWorld.warning("Error occurred when reading file.");
    //         return null;
    //     }
    // }

    /** Returns file size (byte) of source program. */
    public int getFileSize() {
        /* 
         * if (fileSize < 0) {
         *     file2Text();
         * }
         */
        return fileSize;
    }

    /** Returns contents specified offset and length of source program. */
    public String readText(int offset, int length) {
        loadSourceFile();
        try {
            return text.substring(offset, offset + length);
        } catch (Exception e) {
            FrjWorld.error("Invalid offset or length.");
            FrjWorld.error("text length: " + text.length() +
                           ", offset: " + offset + ", length: " + length);
            return null;
        }
    }

    /** Returns raw contents specified offset and length of source program. */
    public byte[] readBytes(int offset, int length) {
        loadSourceFile();
        try {
            return text.getBytes(offset, offset + length);
        } catch (Exception e) {
            FrjWorld.error("Invalid offset or length.");
            FrjWorld.error("text length: " + text.length() +
                           ", offset: " + offset + ", length: " + length);
            return null;
        }
    }


    /**
     * @return encoding string for this source file
     */
    public String getEncoding() {
        return encoding;
    }

    /**
     * Returns line number of specified offset.
     * @return line number, -1 if offset is invalid.
     */
    public int offset2Line(int offset) {
        loadSourceFile();
        if (offset < 0 || getFileSize() < offset) {
            return -1;
        }

        int lineEndType = text.getLineEndType();
        char lineEnd;
        if (lineEndType == ModifiableStringBuffer.LF ||
            lineEndType == ModifiableStringBuffer.CRLF) {
            lineEnd = '\n';
        } else if (lineEndType == ModifiableStringBuffer.CR ||
                 lineEndType == ModifiableStringBuffer.LFCR) {
            lineEnd = '\r';
        } else {
            return -1;          // Unknown line-end type.
        }

        int ln = 1;             // Line no.
        for (int i = 0; i < offset; i++) {
            if (text.charAt(i) == (char)lineEnd) {
                ln++;
            }
        }
        return ln;
    }

    /**
     * Reads file and set contents to text.
     * Any method which use member "text" should call this first.
     */
    private void loadSourceFile() {
        if (text != null) {
            return;
        }

        try {
            File            file = new File(sourceFileName);
            FileInputStream fin = null;
            if (file.exists()) {
                fin = new FileInputStream(file);
            } else {
                // FrjWorld.warning("Can't open " + sourceFileName + ".");
                FrjWorld.warning("\"" + sourceFileName + ": Can't open.");
                return;
            }
            // !! C.f. SPD_LINE_LEN.
            byte            raw[] = new byte[32 * 1024];
            int             rawSize;
            text = new ModifiableStringBuffer();
            switch (FrjWorld.getEncodingStyle()) {
            case FrjWorld.ENC_FILE:
                text.setEncoding(encoding);
                break;
            case FrjWorld.ENC_SYSTEM:     
                text.setEncoding(FrjWorld.getSystemEncoding());
                break;
            }
            while ((rawSize = fin.read(raw)) != -1) {
                text.append(raw, rawSize);
            }
            if (text.length() != fileSize) {
                FrjWorld.warning("\"" + sourceFileName +
                                 "\": Invalid file size.");
            }
        } catch (IOException e) {
            FrjWorld.warning("\"" + sourceFileName + "\": Can't read.");
        } catch (Exception e) {
            FrjWorld.warning("\"" + sourceFileName + "\": Can't read.");
            e.printStackTrace();
        }
    }

    /** Registers changes. */
    void changeText(int offset, int length, String str, FrjObject obj)
        throws SuchChangeIsNotAllowedException {
        SourceChange change = new SourceChange(offset, length, str, obj);

        for (Enumeration e = changes.elements();
             e.hasMoreElements();) {
            SourceChange c = (SourceChange) e.nextElement();
            if (change.isOverlapped(c)) {
                String mes = "Some changes are overlapped";
                throw new SuchChangeIsNotAllowedException(mes + c + ", "
                                                          + change);
            }
        }
        changes.addElement(change);
    }

    /** Changes source programs actually.
     * Changed program is written in a file named <I>filename</I>
     * Object's offset, length, str and other attributes  are also changed.
     *
     * NOTE: This method doesn't write changed object to freja.
     *       You must invoke FrjWorld#save() method to save objects.
     */
    public void flushChanges(String filename) {
        loadSourceFile();

        writeText(sourceFileName + ".bak");
        SortChangesByOffset sorter = new SortChangesByOffset(changes);
        sorter.sort();
        for (Enumeration e = sorter.getEnumeration();
             e.hasMoreElements();) {
            SourceChange c = (SourceChange)e.nextElement();
            text.replace(c.offset, c.length, c.str);
        }
        changeObjectsAttrs();

        writeText(filename);
        if (hasRegisteredChanges()) {
            modified = true;
        }
        clearChanges();
    }

    /** Changes source programs actually.
     * Changed program is written in the original source file.
     * Object's offset, length, str and other attributes  are also changed.
     *
     * NOTE: This method doesn't write changed object to freja.
     *       You must invoke FrjWorld#save() method to save objects.
     */
    public void flushChanges() {
        loadSourceFile();
        flushChanges(sourceFileName);
    }

    /** Clears registered changes. */
    public void clearChanges() {
        changes.removeAllElements();
    }

    /** Returns true if there are registed changes */
    public boolean hasRegisteredChanges() {
        return (changes.size() > 0);
    }

    /**
     * Returns true if a source program is modified.
     *
     * NOTE: When changes are registered and flushed,
     *       a source program is <b>modified</b>.
     */
    public boolean isModified() {
        return modified;
    }

    void changeObjectsAttrs() {
        int offsetDiffs[][] = new int[changes.size() +1][2];
        int OFFSET = 0, DIFF = 1;

        offsetDiffs[0][OFFSET] = offsetDiffs[0][DIFF] = 0; // Sentinel.

        RSortChangesByOffset changeSorter = new RSortChangesByOffset(changes);
        changeSorter.sort();
        int i = 1;
        for (Enumeration cs = changeSorter.getEnumeration();
             cs.hasMoreElements();) {
            SourceChange c = (SourceChange) cs.nextElement();
            FrjObject obj = c.obj;
            int curDiff = c.str.length() - obj.getStrLength();;
            offsetDiffs[i][OFFSET] = c.offset;
            offsetDiffs[i][DIFF] =      offsetDiffs[i-1][DIFF] + curDiff;
            Enumeration objs = getObjectsIncluding(obj.getStrOffset(),
                                                   obj.getStrOffset()
                                                   + obj.getStrLength());
            while (objs.hasMoreElements()) {
                FrjObject o = (FrjObject) objs.nextElement();
                o.setLength(o.getLength() + curDiff);
            }
            obj.setStr(c.str);
            i++;
        }

        Enumeration objs = FrjExtents.getAllObjects(getFrejaID());
        SortByOffset objSorter = new SortByOffset(objs);
        objSorter.sort();
        for (objs = objSorter.getEnumeration();
             objs.hasMoreElements();) {
            FrjObject obj = (FrjObject) objs.nextElement();
            for (i = changes.size(); i >= 0; i--) {
                if (offsetDiffs[i][OFFSET] < obj.getOffset()) {
                    break;
                }
            }
            if (i > 0) {
                obj.setOffset(obj.getOffset() + offsetDiffs[i][DIFF]);
            }
            for (i = changes.size(); i >= 0; i--) {
                if (offsetDiffs[i][OFFSET] < obj.getStrOffset()) {
                    break;
                }
            }
            if (i > 0) {
                obj.setStrOffset(obj.getStrOffset() + offsetDiffs[i][DIFF]);
            }
        }
    }
                
    void writeText(String filename) {
        try {
            File file = new File(filename);
            FileWriter fout = new FileWriter(file);
            fout.write(text.toString());
            fout.flush();
        } catch (IOException e) {
            FrjWorld.error("Can't write file " + filename + ".");
            e.printStackTrace();
        }
    }

    private Enumeration getObjectsIncluding(int start, int end) {
        Vector v = new Vector();
        for (Enumeration e = FrjExtents.getAllObjects(getFrejaID());
             e.hasMoreElements();) {
            FrjObject obj = (FrjObject) e.nextElement();
            if (obj.getOffset() <= start
                && end <= obj.getOffset() + obj.getLength()) {
                v.addElement(obj);
            }
        }
        return v.elements();
    }

    protected void setSourceFileInfo() {
        if (!exists()) {
            return;
        }
        try {
            /* Get source filename from "SDB/XXX/File". */
            DBQuery dbQuery = new DBQuery(this, "File");
            DBClassObject obj;
            if (!dbQuery.trylock4Reading()) {
                return;
            }
            while ((obj = dbQuery.getAClassTuple()) != null) {
                this.sourceFileName = obj.getStringValueByName("str");
                int enc = obj.getIntValueByName("encoding");
                this.encoding = FrjFile.getJavaEncodingStringByInt(enc);
                this.fileSize = obj.getIntValueByName("length");
            }
            dbQuery.unlock4Reading();
        } catch (DBException e) {
            FrjWorld.warning(e.getMessage());
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public String getPackageName() {
        return packageName;
    }

    public int getMaxCRID(int classRelID){
        return maxCRID[SymSort.symSort2TableIndex(classRelID)];
    }

    public int incrementMaxCRID(int classRelID){
        return ++maxCRID[SymSort.symSort2TableIndex(classRelID)];
    }

    public int setMaxCRID(int classRelID, int maxCRIDval){
        maxCRID[SymSort.symSort2TableIndex(classRelID)] = maxCRIDval;
        return getMaxCRID(classRelID);
    }
}

class SourceChange {
    int                 offset;
    int                 length;
    String              str;
    FrjObject           obj;

    SourceChange(int offset, int length, String str, FrjObject obj) {
        this.offset = offset;
        this.length = length;
        this.str = str;
        this.obj = obj;
    }

    boolean isOverlapped(SourceChange c) {
        // if ((this.offset + this.length < c.offset) ||
        //     (c.offset + c.length < this.offset)) {
        //     return false;
        // } else {
        //     return true;
        // }
        return !((this.offset + this.length < c.offset) ||
                 (c.offset + c.length < this.offset));
    }

    public String toString() {
        return "offset: " + offset + " length: " + length + " str: " + str;
    }
}

/**
 * Sorts changes by offset.
 */

class SortChangesByOffset extends VectorSorter {
    public SortChangesByOffset(Vector v) {
        super(v);
    }

    public SortChangesByOffset(Enumeration e) {
        super(e);
    }

    public SortChangesByOffset() {
        super();
    }

    public int compare(Object a, Object b) {
        SourceChange changeA = (SourceChange)a;
        SourceChange changeB = (SourceChange)b;

        // if (changeA.offset < changeB.offset) {
        //     return 1;
        // } else {
        //     return -1;
        // }
        return (changeA.offset < changeB.offset) ? 1 : -1; 
    }
}

/**
 * Sorts changes by reverse offset order.
 */

class RSortChangesByOffset extends VectorSorter {
    public RSortChangesByOffset(Vector v) {
        super(v);
    }

    public RSortChangesByOffset(Enumeration e) {
        super(e);
    }

    public RSortChangesByOffset() {
        super();
    }

    public int compare(Object a, Object b) {
        SourceChange changeA = (SourceChange)a;
        SourceChange changeB = (SourceChange)b;

        // if (changeA.offset > changeB.offset) {
        //     return 1;
        // } else {
        //     return -1;
        // }
        // return (changeA.offset > changeB.offset) ? 1 : -1;
        return (changeB.offset < changeA.offset) ? 1 : -1;
    }
}
