/* 
 * Program:             $RCSfile: FrjClassInterface.java,v $  $Revision: 41.13 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjClassInterface.java,v 41.13 2003/02/12 03:39:31 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

/**
 * Class FrjClassInterface is generalized FrjClass and FrjInterface.
 */

abstract public class FrjClassInterface extends FrjIdentifier {
    // Related objects.
    protected FrjFile           file = null;
    protected Vector            memberClasses = null;
    protected Vector            memberInterfaces = null;
    protected Vector            routines = null;
    protected Vector            members = null;
    protected Vector            superClassInterfaceTypes = null;
    protected Vector            types = null;
    protected FrjClass          outerClass = null;
    protected FrjInterface      outerInterface = null;

    // Constructors.
    public FrjClassInterface(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
    }
    public FrjClassInterface(FrejaEntry entry) {
        super(entry);
    }
    public FrjClassInterface() {
        super();
    }

    public void setRelatedObjects() {
        superClassInterfaceTypes = new Vector();
        types = new Vector();
        super.setRelatedObjects();
        VectorSorter sorter = new SortByOffset(superClassInterfaceTypes);
        sorter.sort();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        superClassInterfaceTypes = null;
        types = null;
    }

    /** Traverses <I>define_class</I> or <I>define_interface</I>. */
    public FrjFile getFile() {
        return file;
    }

    /** Traverses <I>class_class</I> or <I>interface_class</I>. */
    public Enumeration getMemberClasses() {
        return memberClasses.elements();
    }
    /** Traverses <I>class_interface</I> or <I>interface_interface</I>. */
    public Enumeration getMemberInterfaces() {
        return memberInterfaces.elements();
    }

    /** Traverses <I>class_routine</I> or <I>interface_routine</I>. */
    public Enumeration getRoutines() {
        return routines.elements();
    }
    /** Traverses <I>class_variable</I> or <I>interface_variable</I>. */
    public Enumeration getMembers() {
        return members.elements();
    }
    /** Traverses <I>type_class</I> or <I>type_interface</I>. */
    public abstract Enumeration getTypes();

    /** Traverses <I>class_class</I> or <I>interface_class</I>. */
    public FrjClass getOuterClass() {
        return outerClass;
    }
    /** Traverses <I>interface_class</I> or <I>interface_interface</I>. */
    public FrjInterface getOuterInterface() {
        return outerInterface;
    }

    /**
     * Traverses <I>extend_class</I> or <I>implement_interface</I>
     * or <I>extend_interface</I>
     * @return a super class and super interfaces <P>
     * NOTE: Super interfaces are implemented interface by the class
     *       and extended interfaces by the interface.
     */
    public Enumeration getSuperClassInterfaceTypes() {
        return superClassInterfaceTypes.elements();
    }

    /**
     * Searches method (<I>name</I>) for in this class or interface.
     * Method's parameters are spcified argument of <I>args</I>.
     * If FrjWorld.useCreja is set to true, method is searched for
     * super class and implemented interfaces.
     */
    public FrjRoutine search4Method(String name, FrjTypeList args) {
        // Search a routine that has all the same parameter types.
        if (args == null) {
            FrjWorld.vwarning("\"" + name + "\": " +
                             "Can't resolve method (null arguments).");
            return null;
        }
        FrjRoutine r = null;
        FrjTypeList tl = null;
        try {
            for (Enumeration e = getRoutines(); e.hasMoreElements();
                 /* Empty. */) {
                r = (FrjRoutine) e.nextElement();
                if ((r.getStr()).equals(name)) {
                    tl = r.parameter2TypeList();
                    if (args.equals(tl)) {
                        return r;
                    }
                }
            }

            // Search a routine that has compatible parameter types.
            for (Enumeration e = getRoutines(); e.hasMoreElements();
                 /* Empty. */) {
                r = (FrjRoutine) e.nextElement();
                if ((r.getStr()).equals(name)) {
                    tl = r.parameter2TypeList();
                    if (args.isCompatible(tl)) {
                        return r;
                    }
                }
            }
        } catch (NullPointerException e) {
            FrjWorld.error("\"" + name + "\": " +
                           "Can't resolve method (NullPointerException).");
            // FrjWorld.error("    routine0 name: " + name);
            FrjWorld.error("    routine0 args: " + args);
            // FrjWorld.error("    routine1 name: " + r.getStr());
            FrjWorld.error("    routine1 args: " + tl);
            e.printStackTrace();
            return null;
        }

        // Search super class and interface.
        for (Enumeration e = superClassInterfaceTypes.elements();
             e.hasMoreElements();) {
            FrjType superType = (FrjType) e.nextElement();
            FrjClassInterface c = superType.getReferringClassInterface();
            if (c == null) {
                FrjWorld.vwarning("Can't get \"" + this.getStr() +
                                 "\"'s super class \"" +
                                 superType.getStr() + "\".");
                return null;
            }
            r = c.search4Method(name, args);
            if (r != null) {
                return r;
            }
        }

        // Search outer class.
        if (outerClass != null) {
            r = outerClass.search4Method(name, args);
            if (r != null) {
                return r;
            }
        }

        if (!this.getStr().equals("Object")) {
            // Search for the "Object" class that is the parent of all class.
            FrejaEntry entry = Creja.getFrejaEntry("java.lang.Object");
            if (entry == null) {
                FrjWorld.vwarning("Can't get Freja for \"java.lang.Object\".");
                return null;
            } 
            FrjClass objectClass = null;
            if (!FrjExtents.isLoaded(entry)) {
                if (FrjWorld.isAutoLoading()) {
                    FrjWorld.load(entry);
                } else {
                    return null;
                }
            }
            for (Enumeration e =
                 FrjExtents.getObjects(entry.getFrejaID(),
                                       FrjClass.getClassNameStatically());
                 e.hasMoreElements();) {
                FrjClass c = (FrjClass)e.nextElement();
                if (c.getStr().equals("Object")) {
                    objectClass = c;
                    break;
                }
            }
            if (objectClass != null) {
                return objectClass.search4Method(name, args);
            }
        }
        return null;
    }

    /**
     * Searches constructor for in this class.
     * Constructor's parameters are spcified argument of <I>args</I>.
     * If FrjWorld.useCreja is set to true, method is searched for
     * super class and implemented interfaces.
     */
    public FrjRoutine search4Constructor(FrjTypeList args) {
        // Search a routine that has all the same parameter types.
        if (args == null) {
            FrjWorld.vwarning("Can't resolve constructor (null arguments).");
            return null;
        }
        FrjRoutine r = null;
        FrjTypeList tl = null;
        try {
            for (Enumeration e = getRoutines(); e.hasMoreElements();
                 /* Empty. */) {
                r = (FrjRoutine) e.nextElement();
                if (r.getSort() == FrjRoutine.CONSTRUCTOR) {
                    tl = r.parameter2TypeList();
                    if (args.equals(tl)) {
                        return r;
                    }
                }
            }

            // Search a routine that has compatible parameter types.
            for (Enumeration e = getRoutines();
                 e.hasMoreElements();) {
                r = (FrjRoutine) e.nextElement();
                if (r.getSort() == FrjRoutine.CONSTRUCTOR) {
                    tl = r.parameter2TypeList();
                    if (args.isCompatible(tl)) {
                        return r;
                    }
                }
            }
        } catch (NullPointerException e) {
            FrjWorld.error("Can't resolve constructor (NullPointerException).");
            FrjWorld.error("    routine0 args: " + args);
            FrjWorld.error("    routine1 args: " + tl);
            e.printStackTrace();
            return null;
        }

        // search super class and interface
        for (Enumeration e = superClassInterfaceTypes.elements();
             e.hasMoreElements();) {
            FrjType superType = (FrjType) e.nextElement();
            FrjClassInterface c = superType.getReferringClassInterface();
            if (c == null) {
                FrjWorld.vwarning("Can't get \"" + this.getStr() +
                                 "\"'s super class \"" +
                                 superType.getStr() + "\".");
                return null;
            }
            r = c.search4Constructor(args);
            if (r != null) {
                return r;
            }
        }

        if (!this.getStr().equals("Object")) {
            // Search for the "Object" class that is the parent of all class.
            FrejaEntry entry = Creja.getFrejaEntry("java.lang.Object");
            if (entry == null) {
                FrjWorld.vwarning("Can't get Freja for \"java.lang.Object\".");
                return null;
            } 
            FrjClass objectClass = null;
            if (!FrjExtents.isLoaded(entry)) {
                if (FrjWorld.isAutoLoading()) {
                    FrjWorld.load(entry);
                } else {
                    return null;
                }
            }
            for (Enumeration e =
                 FrjExtents.getObjects(entry.getFrejaID(),
                                       FrjClass.getClassNameStatically());
                 e.hasMoreElements();) {
                FrjClass c = (FrjClass) e.nextElement();
                if (c.getStr().equals("Object")) {
                    objectClass = c;
                    break;
                }
            }
            if (objectClass != null) {
                return objectClass.search4Constructor(args);
            }
        }
        return null;
    }


    /**
     * Searches variable (<I>name</I>) for in this class or interface.
     * If FrjWorld.useCreja is set to true, variable is searched for
     * super class and implemented interfaces.
     */
    public FrjVariable search4Member(String name) {
        // Search for variable.
        for (Enumeration e = getMembers();
             e.hasMoreElements();) {
            FrjVariable var = (FrjVariable) e.nextElement();
            if ((var.getStr()).equals(name)) {
                return var;
            }
        }
        
        // Search super class and interface.
        for (Enumeration e = superClassInterfaceTypes.elements();
             e.hasMoreElements();) {
            FrjType superType = (FrjType) e.nextElement();
            FrjClassInterface c = superType.getReferringClassInterface();
            if (c == null) {
                FrjWorld.vwarning("Can't get \"" + this.getStr() +
                                 "\"'s super class \"" +
                                 superType.getStr() + "\".");
                return null;
            }
            FrjVariable var = c.search4Member(name);
            if (var != null) {
                return var;
            }
        }

        if (!this.getStr().equals("Object")) {
            // Search for the "Object" class that is the parent of all class.
            FrejaEntry entry = Creja.getFrejaEntry("java.lang.Object");
            if (entry == null) {
                FrjWorld.warning("Can't get Freja for \"java.lang.Object\".");
                return null;
            }
            FrjClass objectClass = null;
            if (!FrjExtents.isLoaded(entry)) {
                if (FrjWorld.isAutoLoading()) {
                    FrjWorld.load(entry);
                } else {
                    return null;
                }
            }
            for (Enumeration e = FrjExtents.getObjects(entry.getFrejaID(),
                                                       FrjClass.getClassNameStatically());
                 e.hasMoreElements();) {
                FrjClass c = (FrjClass) e.nextElement();
                if (c.getStr().equals("Object")) {
                    objectClass = c;
                    break;
                }
            }
            if (objectClass != null) {
                return objectClass.search4Member(name);
            }
        }
        return null;
    }

    /**
     * Returns full qualified class name.
     * If package isn't defined, it equals to str.
     * e.g. Object -> java.lang.Object
     */
    public String getFullClassName() {
        if (file == null) {
            return getStr();
        }
        FrjQName pkg = file.getDefinedPackage();
        if (pkg == null) {
            return getStr();
        } else {
            return pkg.getFullString() + "." + getStr();
        }
    }
}
