/* 
 * Program:             $RCSfile: FrjFile.java,v $  $Revision: 41.10 $
 * 
 * Purpose:             The class corresponding to File.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjFile.java,v 41.10 2003/04/02 13:02:07 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjFile extends FrjObject {
    // Class variables
    /** A class name of this class. */
    public final static String  CLASS_NAME = "File";

    /**
     * Table from a encoding to string. Initialized by static block.
     */
    protected static String     encodingNameTable[];
    protected static String     javaEncodingNameTable[];

    // Attribute.
    public int                  encoding;

    // Related objects.
    protected Vector            classes = null;
    protected Vector            interfaces = null;
    protected FrjQName          definedPackage = null;
    protected Vector            importedPackages = null;

    // Constatnts.
    public final static int
        EC_UNKNOWN = 0,
        EC_ASCII = 1,
        EC_JIS = 2,
        EC_EUC = 3,
        EC_EUCORSJIS = 4,
        EC_SJIS = 5;
    protected final static int
        EC_FIRST_MEMBER = EC_UNKNOWN,
        EC_LAST_MEMBER = EC_SJIS;

    // Constructors.
    public FrjFile(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // encoding = EC_UNKNOWN;
    }

    public FrjFile(FrejaEntry entry) {
        super(entry);
        encoding = EC_UNKNOWN;
    }

    public FrjFile() {
        super();
        encoding = EC_UNKNOWN;
    }

    // Class methods.
    /** Return the class name of this object. */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** Create object from SDB. */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjFile(entry, obj);
    }

    // Instance method.
    /** Return the class name dynamically. */
    public String getClassName() {
        return CLASS_NAME;
    }

    /** Return the encoding. */
    public int getEncoding() {
        return encoding;
    }

    /** Check whether this object has a valid encoding or not. */
    public boolean hasValidEncoding() {
        return (EC_FIRST_MEMBER <= encoding && encoding <= EC_LAST_MEMBER);
    }
    /** Return the sort as string. */
    public String getEncodingString() {
        return (hasValidEncoding() ? encodingNameTable[encoding] :
                "INVALID ENCODING");
    }
    /** Return the sort as string. */
    public String getJavaEncodingString() {
        return (hasValidEncoding() ? javaEncodingNameTable[encoding] :
                "INVALID ENCODING");
    }

    /** Return the sort as string. */
    protected static String getJavaEncodingStringByInt(int enc) {
        return ((EC_FIRST_MEMBER <= enc && enc <= EC_LAST_MEMBER) ?
            javaEncodingNameTable[enc] : "INVALID ENCODING");
    }

    // Methods about relation, not implemented yet.
    public void setRelatedObjects() {
        classes = new Vector();
        interfaces = new Vector();
        importedPackages = new Vector();

        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        classes = null;
        interfaces = null;
        definedPackage = null;
        importedPackages = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.DEFINE_CLASS)) {
            classes.addElement((FrjClass) obj);
        }
        else if (relName.equals(FrjRelations.DEFINE_INTERFACE)) {
            interfaces.addElement((FrjInterface) obj);
        }
        else if (relName.equals(FrjRelations.DEFINE_PACKAGE)) {
            definedPackage = (FrjQName) obj;
        }
        else if (relName.equals(FrjRelations.IMPORT_PACKAGE)) {
            importedPackages.addElement((FrjQName) obj);
        }
        else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        }
        else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }

    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(classes);
        sorter.sort();
        sorter.setVector(interfaces);
        sorter.sort();
        sorter.setVector(importedPackages);
        sorter.sort();
    }

    /** Traverses <I>define_class</I>. */
    public Enumeration getClasses() {
        return classes.elements();
    }
    /** Traverses <I>define_interface</I>. */
    public Enumeration getInterfaces() {
        return interfaces.elements();
    }
    /** Traverses <I>define_package</I>. */
    public FrjQName getDefinedPackage() {
        return definedPackage;
    }
    /** Traverses <I>import_package</I>. */
    public Enumeration getImportedPackages() {
        return importedPackages.elements();
    }
        
    /** Return parent object if exists. */
    public FrjObject getParentObject() {
        return null;
    }

    // Traverse method.
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        if (definedPackage != null) {
            definedPackage.traverse(tm);
        }
        traverseObjects(getImportedPackages(), tm);

        tm.doMyself(this);

        traverseObjects(getClasses(), tm);
        traverseObjects(getInterfaces(), tm);   
        tm.doAfterTraverse(this);
    }

    static {
        encodingNameTable = new String[EC_LAST_MEMBER + 1];

        encodingNameTable[EC_UNKNOWN] = "Unknown encoding";
        encodingNameTable[EC_ASCII] = "ASCII";
        encodingNameTable[EC_JIS] = "ISO-2022-JP";
        encodingNameTable[EC_EUC] = "EUC-JP";
        // encodingNameTable[EC_EUCORSJIS] = "EUC or Shift JIS";
        encodingNameTable[EC_EUCORSJIS] = "ISO-8859-1";
        encodingNameTable[EC_SJIS] = "Shift_JIS";

        javaEncodingNameTable = new String[EC_LAST_MEMBER + 1];

        javaEncodingNameTable[EC_UNKNOWN] = "Unknown encoding";
        javaEncodingNameTable[EC_ASCII] = "ASCII";
        javaEncodingNameTable[EC_JIS] = "ISO2022JP";
        javaEncodingNameTable[EC_EUC] = "EUC_JP";
        javaEncodingNameTable[EC_EUCORSJIS] = "ISO8859_1";
        javaEncodingNameTable[EC_SJIS] = "SJIS";
    }
}
