/* 
 * Program:             $RCSfile: FrjLabel.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to Expression.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjLabel.java,v 41.5 2003/02/12 03:39:33 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjLabel extends FrjObject {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Label";

    /** Table from a sort to string. */
    // Initialized by static block.
    protected static String     sortNameTable[];

    // Attribute.
    public int                          sort;

    // Related objects.
    protected FrjLabel          referree = null;
    protected Vector            referrers = null;
    protected FrjStatement      statement = null;
    protected FrjExpression     caseValue = null;

    // Constatnts.
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int        _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int     DEFINED = _i++;
    public final static int     REFERRED = _i++;
    public final static int     CASE = _i++;
    public final static int     DEFAULT = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // sentinel
                
    // Constructors.
    public FrjLabel(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }
    public FrjLabel(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }
    public FrjLabel() {
        super();
        sort = INVALID_ID;
    }

    // Class methods.
    /** Return the class name of this object statically */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjLabel(entry, obj);
    }

    // Instance methods
    /** Return the class name of this object */
    public String getClassName() {
        return CLASS_NAME;
    }

    // About attribute.
    /** Return the sort */
    public int getSort() {
        return sort;
    }
    /** Check this object has a valid sort. */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** Return the sort as string. */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return "INVALID SORT";
        }
        return sortNameTable[sort];
    }

    /** Return the name of attribute and its value as string. */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** Convert object to string. */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // About relation.
    public void setRelatedObjects() {
        referrers = new Vector();
        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        referree = null;
        referrers = null;
        statement = null;
        caseValue = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.STATEMENT_LABEL)) {
            statement = (FrjStatement) obj;
        } else if (relName.equals(FrjRelations.REFER_LABEL)) {
            if (rel.get1stObject().getID() == this.getID()) {
                // Argument obj is a referree.
                referree = (FrjLabel) obj;
            } else {
                // Argument obj is a referrer.
                referrers.addElement(obj);
            }
        } else if (relName.equals(FrjRelations.CASE_VALUE)) {
            caseValue = (FrjExpression) obj;
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(referrers);
        sorter.sort();
    }

    /** Traverses <I>statement_label</I>.*/
    public FrjStatement getStatement() {
        return statement;
    }
    /** Traverses <I>case_value</I>.*/
    public FrjExpression getCaseValue() {
        return caseValue;
    }
    /**
     * Traverses <I>refer_label</I>.
     * @return an instance of FrjLabel that this label refers to
     */
    public FrjLabel getReferree() {
        return referree;
    }
    /**
     * Traverses <I>refer_label</I>.
     * @return instances of FrjLabel that refer to this label
     */
    public Enumeration getReferrers() {
        return referrers.elements();
    }

    public FrjObject getParentObject() {
        return statement;
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjLabel) obj).sort = sort;
    }

    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        tm.doMyself(this);

        if (caseValue != null) {
            caseValue.traverse(tm);
        }

        tm.doAfterTraverse(this);
    }
        
    // static initializer
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];
                
        sortNameTable[DEFINED] = "Defined Label";
        sortNameTable[REFERRED] = "Referred Label";
        sortNameTable[CASE] = "Case Label";
        sortNameTable[DEFAULT] = "Default Label";
    }
}
