/* 
 * Program:             $RCSfile: FrjObject.java,v $  $Revision: 41.9 $
 * 
 * Purpose:             The super class of all the DBClasses.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjObject.java,v 41.9 2003/02/12 03:39:34 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public abstract class FrjObject implements Loadable {
    // Class variables.
    /** A class name of this class. It must be overridden in sub class.  */
    public final static String  CLASS_NAME = "Object";

    /** An entry of the Freja. It is used on getting text */
    protected FrejaEntry        entry = null;

    // Common attributes.
    public long                 id;
    public int                  offset;
    public int                  length;
    public int                  strOffset;
    public String               str;

    // Relations.
    protected FrjRelations      relations;
    protected Vector            comments;

    // Constants.
    /** invalid id, offset, length */
    protected final static int INVALID_ID = -1;
    protected final static int  MAX_ATTR = 10;
    protected final static int
        ID_INDEX = 0,
        OFFSET_INDEX = 1,
        LENGTH_INDEX = 2,
        STROFFSET_INDEX = 3,
        STR_INDEX = 4,
        ORIGINAL_ATTR_START = 5;
        
    // Constructors 
    public FrjObject(FrejaEntry entry, DBClassObject obj) {
        /* set values */
        for (int i = 0; i < obj.getAttrNum(); i++) {
            if (obj.attrIsIntAt(i)) {
                if (obj.getAttrNameAt(i).equals("id")) {
                    long newId = FrjID.generateObjectID(entry, obj.getIntValueAt(i));
                    ReflectionUtil.setAttrValLongByName(this, obj.getAttrNameAt(i), newId);
                } else {
                    ReflectionUtil.setAttrValIntByName(this, obj.getAttrNameAt(i), obj.getIntValueAt(i));
                }
            } else if (obj.attrIsStringAt(i)) {
                ReflectionUtil.setAttrValStringByName(this, obj.getAttrNameAt(i), obj.getStringValueAt(i));
            } else {
                /* Nothing. */
            }
        }
        
        relations = new FrjRelations();
        this.entry = entry;
        
        FrjExtents.addObject(this);
    }

    /**
     * Creates a new object (not from SDB).
     * ID is set as object in entry and relations is allocated,
     * but offset, length, string is not set (set to INVALID_ID, null).
     * Created object is added FrjExtents.
     */
    public FrjObject(FrejaEntry entry) {
        this.entry = entry;
        id = FrjID.generateObjectID(entry, getClassName());
        offset = INVALID_ID;
        length = INVALID_ID;
        strOffset = INVALID_ID;
        this.str = null;
        relations = new FrjRelations();

        FrjExtents.addObject(this);
    }

    /**
     * Creates a new object (not from SDB).
     * ID is set correctly and relations is allocated,
     * but offset, length, string is not set (set to INVALID_ID, null).
     * Created object is NOT added FrjExtents.
     */
    public FrjObject() {
        id = FrjID.generateObjectID();
        offset = INVALID_ID;
        length = INVALID_ID;
        strOffset = INVALID_ID;
        this.str = null;
        relations = new FrjRelations();
    }

    /**
     * Returns the class name statically,
     * this method must be overridden in a sub class.
     */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** @return: <I>s</I> is null string? */
    public static boolean isNullString(String s) {
        return s.equals("");
    }
        
    // Instance methods
    /** @return the ID */
    public long getID() {
        return id;
    }
    /** @return the offset of this object */
    public int getOffset() {
        return offset;
    }
    /** @return the length of this object */
    public int getLength() {
        return length;
    }
    /** @return the offset of str */
    public int getStrOffset() {
        return strOffset;
    }
    /** @return the length of str */
    public int getStrLength() {
        return str.length();
    }

    /** @return the str */
    public String getStr() {
        return str;
    }

    /** Sets offset. */
    protected void setOffset(int offset) {
        this.offset = offset;
    }
    /** Sets length. */
    protected void setLength(int length) {
        this.length = length;
    }
    /** Sets strOffset. */
    protected void setStrOffset(int offset) {
        this.strOffset = offset;
    }
    /** Sets str. */
    protected void setStr(String str) {
        this.str = str;
    }

    /** @return comments */
    public Enumeration getComments() {
        return comments.elements();
    }
    /**
     * Returns the first line number of this object.
     * @return          line number, -1 if offset is invalid
     */
    public int getLineNo() {
        return entry.offset2Line(offset);
    }

    /** Checks whether this object has a valid ID. */
    public boolean hasValidID() {
        return (id != INVALID_ID);
    }
    /** Checks whether this object has valid offset. */
    public boolean hasValidOffset() {
        return (offset >= 0);
    }
    /** Checks wheter this object has valid length. */
    public boolean hasValidLength() {
        return (length >= 0);
    }
    /** Checks whether this object has valid strOffset. */
    public boolean hasValidStrOffset() {
        return (strOffset >= 0);
    }

    /** Returns the id as String. */
    public String getIDString() {
        return (hasValidID()) ? Long.toString(id) : "INVALID ID";
    }
    /** Returns the offset as String. */
    public String getOffsetString() {
        return (hasValidOffset())
            ? Integer.toString(offset) : "INVALID OFFSET";
    }
    /** Returns the length as String. */
    public String getLengthString() {
        return (hasValidLength())
            ? Integer.toString(length) : "INVALID LENGTH";
    }
    /** Returns the strOffset as String. */
    public String getStrOffsetString() {
        return (hasValidStrOffset())
            ? Integer.toString(strOffset) : "INVALID STROFFSET";
    }

    /** Return the name of attributes and its value as String. */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = new FrjAttrsAsString[MAX_ATTR];
                
        attrs[ID_INDEX] = new FrjAttrsAsString("ID", getIDString());
        attrs[OFFSET_INDEX]
            = new FrjAttrsAsString("offset", getOffsetString());
        attrs[LENGTH_INDEX]
            = new FrjAttrsAsString("length", getLengthString());
        attrs[STROFFSET_INDEX]
            = new FrjAttrsAsString("strOffset", getStrOffsetString());
        attrs[STR_INDEX] = new FrjAttrsAsString("str", str);
                
        for (int i = STR_INDEX+1; i < MAX_ATTR; i++) {
            attrs[i] = null;
        }

        return attrs;
    }

    /** Converts to String. */
    public String toString() {
        String res;
                
        res = "[" + getClassName() + "] ";
        res += "ID: " + getIDString();
        res += ", offset: " + getOffsetString();
        res += ", length: "+ getLengthString();
        res += ", strOffset: " + getStrOffsetString();
        res += ", str: " + str;
        return res;
    }

    /** Returns FrejaEntry of this object. */
    public FrejaEntry getFrejaEntry() {
        return entry;
    }

    // Abstract methods.
    /** Returns class name dynamically. */
    public String getClassName() {
        return CLASS_NAME;
    }

    /** Creates new objects from SDB */
    abstract protected void create(FrejaEntry entry, DBClassObject obj);

    /** Sets related objects from relations. */
    public void setRelatedObjects() {
        comments = new Vector();
        for (Enumeration e = getRelations();
             e.hasMoreElements();) {
            FrjRelation rel = (FrjRelation) e.nextElement();
            FrjObject obj = rel.getRelatedObject(this);
            try {
                setRelatedObject(obj, rel);
            } catch (NoSuchRelatedObjectException ext) {
                ext.printStackTrace();
            }
        }
    }

    /** Resets related objects. */
    public void resetRelatedObjects() {
        FrjRelations rels = relations;
        removeRelatedObjects();
        relations = rels;
        setRelatedObjects();
    }

    /** Removes all related objects. */
    public void removeRelatedObjects() {
        relations = new FrjRelations();
        // relations.removeAllElement() can't be used
        // because resetRelatedObject() reserves relations.
        comments = null;
    }

    /** Removes specified related object. */
    public void removeRelatedObject(FrjObject obj)
        throws NoSuchRelatedObjectException {
        FrjRelation rel = getRelation(obj);
        if (rel == null) {
            throw new NoSuchRelatedObjectException(obj + " is not related.");
        }
        relations.removeElement(rel);
        rel.unlinkRelatedObject();
        resetRelatedObjects();
    }

    /** Removes this object. */
    public void remove() {
        try {
            for (Enumeration e = getRelatedObjects();
                 e.hasMoreElements();) {
                FrjObject obj = (FrjObject) e.nextElement();
                obj.removeRelatedObject(this);
                this.removeRelatedObject(obj);
            }
        } catch (NoSuchRelatedObjectException e) {
            if (FrjWorld.isVerbose()) {
                FrjWorld.warning("FrjObject#remove(): Already removed? "
                                        + e.getMessage());
            }
        }
        FrjExtents.removeObject(this);
    }   

    /** Sets a related object from <I>object</I> and <I>rel</I>. */
    abstract protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException;

    protected void setRelatedComment(FrjObject obj) {
        if (!(this instanceof FrjComment)) {
            // Add comment.
            comments.addElement(obj);
        }
        // If this is instace of FrjComment,
        // setRelatedComment() is overrided in FrjComment.
    }

    /** Returns all relations as Enumerations. */
    public Enumeration getRelations() {
        return relations.elements();
    }

    /** Returns a relation that relates <I>obj</I>. */
    public FrjRelation getRelation(FrjObject obj) {
        for (Enumeration e = getRelations();
             e.hasMoreElements();) {
            FrjRelation rel = (FrjRelation) e.nextElement();
            if (rel.getRelatedObject(this).getID() == obj.getID()) {
                return rel;
            }
        }
        return null;    /* this statement can't be reached */
    }

    /** Return all related objects as Enumerations. */
    public Enumeration getRelatedObjects() {
        return relations.getRelatedObjects(this);
    }

    /** Sorts related objects by <I>sorter</I>. */
    public abstract void sortRelatedObjects(VectorSorter sorter);


    /** Returns the parent object if exists. */
    abstract public FrjObject getParentObject();

    /**
     * Returns the youngest ancestor of ancestors being instance of className.
     * @param className ancestor's class name
     * @return FrjObject that is ancestor of this object,
     *         null if don't have any ancestor.
     */
    public FrjObject getAncestorObject(String className) {
        FrjObject obj = this;

        while (obj != null && !obj.getClassName().equals(className)) {
            obj = obj.getParentObject();
        }
        return obj;
    }

    /**
     * @return FrjClass that defines this object, null if dont't have.
     */
    public FrjClassInterface getDefinerClassInterface() {
        FrjObject obj = this;

        while (obj != null && ! (obj instanceof FrjClassInterface)) {
            obj = obj.getParentObject();
        }
        return (FrjClassInterface) obj;
    }

    /** Copies object's attributes and relations to <I>obj</I>. */
    public void copyTo(FrjObject obj) {
        obj.id = this.id;
        obj.offset = this.offset;
        obj.length = this.length;
        obj.strOffset = this.strOffset;
        obj.str = this.str;

        obj.relations = this.relations;
        obj.comments = this.comments;
        obj.entry = this.entry;
    }

    /** Copies object's attributes (no included relations) to <I>obj</I>. */
    /*
     * public void copyAttrTo(FrjObject obj) {
     * obj.id = this.id;
     * obj.str = this.str;
     * obj.offset = this.offset;
     * obj.length = this.length;
     * }
     */

    /**
     * Returns text from offset to length.
     * <P>
     * NOTE: getText() differs from getStr().
     */
    public String getText() {
        return entry.readText(offset, length);
    }

    /**
     * If this is the same object as <I>obj</I>,
     * i.e. has the same id, return true.
     */
    public boolean isSame(FrjObject obj) {
        return (this.getID() == obj.getID());
    }

    /**
     * Changes str of this object to <I>s</I>.
     * If str equals to null, throws SuchChangeIsNotAllowedException.
     */
    protected void changeToken(String s)
        throws SuchChangeIsNotAllowedException {
        if (str == null) {
            String mes = "This object doesn't have str attribute.\n  ";
            throw new SuchChangeIsNotAllowedException(mes + this);
        }
        /*
          System.err.println("offset: " + offset + " length: " + length
          + " str: " + entry.readText(offset, length)
          + " -> " + s);
        */
        entry.changeText(getStrOffset(), getStrLength(), s, this);
    }
                        
                

    /**
     * Traverses child objects.
     */
    abstract public void traverse(TraverseMethod tm);
    protected void traverseObjects(Enumeration e, TraverseMethod tm) {
        while (e.hasMoreElements()) {
            ((FrjObject) e.nextElement()).traverse(tm);
        }
    }

    /**
     * Loads objects from SDB.
     *
     * <P>
     * example:<P>
     * <PRE>
     *   Class c = Class.forName("Japid.Freja.Frj" + className);
     *   Loadable loader = (Loadable) c.newInstance();
     *   loader.load(entry, className);
     * </PRE>
     */
    public void load(FrejaEntry entry, String name) {
        DBClassObject   obj;

        DBQuery dbQuery = new DBQuery(entry, getClassName());
        try {
            if (!dbQuery.trylock4Reading()) {
                return;
            }
            while ((obj = dbQuery.getAClassTuple()) != null) {
                this.create(entry, obj);
            }
            dbQuery.unlock4Reading();
        } catch (DBException e) {
            FrjWorld.warning(e.getMessage());
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Saves objects from SDB.
     *
     * <P>
     * example:<P>
     * <PRE>
     *   Class c = Class.forName("Japid.Freja.Frj" + className);
     *   Savable saver = (Savable) c.newInstance();
     *   saver.save(entry, className);
     * </PRE>
     */
    public void save(FrejaEntry entry, String name) {
        DBQuery dbQuery = new DBQuery(entry, getClassName());
        try {
            if (!dbQuery.trylock4Writing()) {
                return;
            }
                        
            VectorSorter sorter
                = new SortByID(FrjExtents.getObjects(entry.getFrejaID(),
                                                     getClassName()));
            sorter.sort();
            for (Enumeration e = sorter.getEnumeration();
                 e.hasMoreElements();) {
                FrjObject obj = (FrjObject) e.nextElement();
                dbQuery.writeAClassTuple(DBClassObject.getDBClassObject(obj));
            }
            dbQuery.unlock4Writing();
        } catch (DBException e) {
            FrjWorld.warning(e.getMessage());
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    protected String toSdbline() {
        String res = "";
        res += FrjID.getFrejaDataID(getID()) + JDBConsts.SDB_ATTR_SEPARATOR;
        res += getOffset() + JDBConsts.SDB_ATTR_SEPARATOR;
        res += getLength() + JDBConsts.SDB_ATTR_SEPARATOR;
        res += getStrOffset() + JDBConsts.SDB_ATTR_SEPARATOR;
        res += stringToSdbformat(getStr()) + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    protected String stringToSdbformat(String s) {
        // if (s == null || isNullString(s)) {
        //     return "";
        // } else {
        //     return s;
        // }
        return ((s == null || isNullString(s)) ? "" : s);
    }

    /**
     * Sets comment.<P>
     */
    public void setComment(FrjComment com) {
        FrjRelation rel = new FrjRelation(FrjRelations.SYMBOL_COMMENT,
                                          this, com);
        try {
            this.setRelatedObject(com, rel);
            com.setRelatedObject(this, rel);
        } catch (NoSuchRelatedObjectException e) {
            FrjWorld.warning(e.getMessage());
        }
    }
}
