/* 
 * Program:             $RCSfile: FrjOperator.java,v $  $Revision: 41.6 $
 * 
 * Purpose:             The class corresponding to Expression.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjOperator.java,v 41.6 2003/02/12 03:39:35 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjOperator extends FrjObject {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Operator";

    /** Table from a sort to string. */
    // initialized by static block.
    protected static String     sortNameTable[];

    // Attribute.
    public int                  sort;

    // Related objects.
    protected FrjExpression     expression = null;
    protected FrjQName          qname = null;
        
    // Constatnts.
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int        _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i;         // Sentinel.
    public final static int     PREFIX = _i++;
    public final static int     POSTFIX = _i++;
    public final static int     BINARY = _i++;
    public final static int     TERNARY = _i++;
    public final static int     PAREN = _i++;
    public final static int     CAST = _i++;
    public final static int     NEW_ARRAY = _i++;
    public final static int     NEW_OBJECT = _i++;
    public final static int     ALLOCATE_DIM = _i++;
    public final static int     REFER_DIM = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // Sentinel.

    protected static int        _j = 0;
    public final static int     LEFT_RIGHT = _j++;
    public final static int     RIGHT_LEFT = _j++;
    public final static int     UNKNOWN = _j++;
    public final static int     NO_SENSE = _j++;

    // Constructors.
    public FrjOperator(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }
    public FrjOperator(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }
    public FrjOperator() {
        super();
        sort = INVALID_ID;
    }

    /** Return the class name of this object statically. */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** Create object from SDB. */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjOperator(entry, obj);
    }

    // Instance methods.
    /** Return the class name of this object. */
    public String getClassName() {
        return CLASS_NAME;
    }

    // About attribute
    /** Return the sort. */
    public int getSort() {
        return sort;
    }
    /** Check this object has a valid sort. */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** Return the sort as string. */
    public String getSortString() {
        if (!hasValidSort()) {  /* error. */
            return null;
        }
        return sortNameTable[sort];
    }

    /** Return the name of attribute and its value as string. */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** Returns true if an assignment operator. */
    public boolean isAssignment() {
        if (str.length() == 1) {
            return str.equals("=");
        } else {
            return (str.endsWith("=")
                    && !(str.equals("==") || str.equals("!=")
                         || str.equals("<=") || str.equals(">=")));
            // +=, *=, <<=, ... and so on
            // exclude ==, !=, <=, >=
        }
    }

    /**
     * Returns true if a conditional operator.
     * Conditional operators are <P>
     *  <PRE> ==, <, >, <=, >=, !, !=, &&, ||, instanceof </PRE>.
     * It doesn't mean <PRE>?:</PRE>.
     * NOTE: If this operator is bitwise and logical operator,
     *  (&, ^, |) returns false.
     * @return  true if a conditional operator (except logical operator)
     */
    public boolean isConditional() {
        return (str.equals("==") || str.equals("<") || str.equals(">")
                || str.equals("<=") || str.equals(">=")|| str.startsWith("!")
                || str.equals("&&") || str.equals("||")
                || str.equals("instanceof"));
    }

    /**
     * Returns true if a bitwise(logical) operator (&, ^, |).
     */
    public boolean isBitwise() {
        return (str.equals("&") || str.equals("^") || str.equals("|"));
    }

    /** Returns true if a dot operator. */
    public boolean isDot() {
        return str.equals(".");
    }

    /** Returns priority of evaluating operand.  */
    public int getOperandPriority() {
        if (getSort() == BINARY) {
            if (isAssignment()) {
                return RIGHT_LEFT;
            } else {
                return LEFT_RIGHT;
            }
        } else if (getSort() == TERNARY) {
            return LEFT_RIGHT;
        } else if (getSort() == REFER_DIM) {
            return RIGHT_LEFT;
        } else {
            return NO_SENSE;
        }
    }

    /** Convert object to string. */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // About relation.
    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        expression = null;
        qname = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.EXPRESSION_OPERATOR)) {
            expression = (FrjExpression) obj;
        } else if (relName.equals(FrjRelations.QNAME_DOT)) {
            qname = (FrjQName) obj;
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    public void sortRelatedObjects(VectorSorter sorter) {
    }

    /** Traverses <I>expression_operator</I>. */
    public FrjExpression getExpression() {
        return expression;
    }
    /** Traverses <I>qname_dot</I>. */
    public FrjQName getQName() {
        return qname;
    }
    public FrjObject getParentObject() {
        // return (expression != null) ? expression : qname;
        if (expression != null) {
            return expression;
        } else {
            return qname;
        }
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjOperator)obj).sort = sort;
    }
                
    // Traverse method.
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);
        tm.doMyself(this);
        tm.doAfterTraverse(this);
    }
        
    // Static initializer.
    static {
        sortNameTable = new String[SORT_LAST_MEMBER + 1];
                
        sortNameTable[PREFIX] = "Prefix";
        sortNameTable[POSTFIX] = "Postfix";
        sortNameTable[BINARY] = "Binary";
        sortNameTable[TERNARY] = "Ternary";
        sortNameTable[PAREN] = "Paren";
        sortNameTable[CAST] = "Cast";
        sortNameTable[NEW_ARRAY] = "New Array";
        sortNameTable[NEW_OBJECT] = "New Object";
        sortNameTable[ALLOCATE_DIM] = "Allocate Dimension";
        sortNameTable[REFER_DIM] = "Refer Dimension";
    }
}
