/* 
 * Program:             $RCSfile: FrjRoutine.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to Statement.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjRoutine.java,v 41.5 2003/02/15 11:56:29 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjRoutine extends FrjIdentifier {
    // Class variables
    /** A class name of this class */
    public final static String  CLASS_NAME = "Routine";

    /** table from a sort to string */
    protected static String     sortNameTable[];        // initialized by static block
        
    // Attributes
    public int                  sort;

    // Related objects
    protected FrjClassInterface         definedClassInterface = null;
    //  protected Vector                modifiers = null;
    protected FrjType           returnType = null;
    protected Vector            parameters = null;
    protected Vector            throwTypes = null;
    protected FrjStatement      body = null;
    protected Vector            callers = null;

    // Constatnts
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int        _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int             METHOD = _i++;
    public final static int             CONSTRUCTOR = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;         // sentinel

    // Constructors
    public FrjRoutine(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        //this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }
                
    public FrjRoutine(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }

    public FrjRoutine() {
        super();
        sort = INVALID_ID;
    }


    // Class methods
    /** return the class name of this object statically */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }
                        
    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjRoutine(entry, obj);
    }

    // Instance methods
    /** return the class name of this object */
    public String getClassName() {
        return CLASS_NAME;
    }

    // about attributes
    /** return the sort */
    public int getSort() {
        return sort;
    }
    /** check this object has a valid sort */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** return the sort as string */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return "INVALID SORT";
        }
        return sortNameTable[sort];
    }

    /** return the name of attribute and its value as string */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** convert object to string */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // about relations
    public void setRelatedObjects() {
        modifiers = new Vector();
        parameters = new Vector();
        throwTypes = new Vector();
        callers = new Vector();

        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        definedClassInterface = null;
        modifiers = null;
        returnType = null;
        parameters = null;
        throwTypes = null;
        body = null;
        callers = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.CLASS_ROUTINE)
            || relName.equals(FrjRelations.INTERFACE_ROUTINE)) {
            definedClassInterface = (FrjClassInterface) obj;
        }
        else if (relName.equals(FrjRelations.ROUTINE_MODIFIER)) {
            modifiers.addElement((FrjModifier) obj);
        }
        else if (relName.equals(FrjRelations.RETURN_TYPE)) {
            returnType = (FrjType) obj;
        }
        else if (relName.equals(FrjRelations.THROW_TYPE)) {
            throwTypes.addElement((FrjType) obj);
        }
        else if (relName.equals(FrjRelations.PARAMETER_VARIABLE)) {
            parameters.addElement((FrjVariable) obj);
        }
        else if (relName.equals(FrjRelations.ROUTINE_BODY)) {
            body = (FrjStatement) obj;
        }
        else if (relName.equals(FrjRelations.INVOKE_ROUTINE)) {
            callers.addElement(obj);
        }
        else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        }
        else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(modifiers);
        sorter.sort();
        sorter.setVector(parameters);
        sorter.sort();
        sorter.setVector(throwTypes);
        sorter.sort();
        sorter.setVector(callers);
        sorter.sort();
    }

    /** Traverses <I>class_routine</I> or <I>interface_routine</I>. */
    public FrjClassInterface getDefinedClassInterface() {
        return definedClassInterface;
    }
    /** Traverses <I>return_type</I>. */
    public FrjType getReturnType() {
        return returnType;
    }
    /**
     * Traverses <I>parameter_variable</I>.
     * @return parameters (FrjVariable)
     */
    public Enumeration getParameters() {
        return parameters.elements();
    }
    /**
     * Traverses <I>throw_type</I>.
     * @return thrown types (FrjType)
     */
    public Enumeration getThrowTypes() {
        return throwTypes.elements();
    }
    /** Traverses <I>routine_body</I>. */
    public FrjStatement getBody() {
        return body;
    }
    /**
     * Traverses <I>invoke_routine</I>.
     * @return expression which invoke this routine<P>
     * NOTE: This doesn't return all invoked expressions.
     */
    public Enumeration getCallers() {
        return callers.elements();
    }

    /**
     * Converts parameters to the list of type.
     * @return the list of type (FrjTypeList)
     */
    public FrjTypeList parameter2TypeList() {
        FrjTypeList             typeList = new FrjTypeList();
        VectorSorter    sorter = new SortByOffset(getParameters());

        sorter.sort();
        for (Enumeration e = sorter.getVector().elements();
             e.hasMoreElements();) {
            FrjVariable param = (FrjVariable) e.nextElement();
            typeList.addType((FrjType) param.getType());
        }
        return typeList;
    }

    /** Returns true if this and parameter routine has the same signature. */
    public boolean isSameSignature(FrjRoutine r) {
        if (this.getStr().equals(r.getStr())) { // same name
            if (returnType.equals(r.getReturnType())) { // same return type
                VectorSorter sorter = new SortByOffset();
                sortRelatedObjects(sorter);
                r.sortRelatedObjects(sorter);
                Enumeration thisParams = getParameters();
                Enumeration rParams = r.getParameters();
                while (thisParams.hasMoreElements()) {
                    if (!rParams.hasMoreElements()) {
                        return false;
                    }
                    FrjVariable thisParam
                        = (FrjVariable) thisParams.nextElement();
                    FrjType thisPType = thisParam.getType();
                    FrjVariable rParam
                        = (FrjVariable) rParams.nextElement();
                    FrjType rPType = rParam.getType();
                    if (!thisPType.equals(rPType)) {
                        return false;
                    }
                }
                if (rParams.hasMoreElements()) {
                    return false;
                }
                return true;
            }
        }
        return false;
    }

    /** return the parent object */
    public FrjObject getParentObject() {
        return definedClassInterface;
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjRoutine) obj).sort = sort;
    }

    /** Returns signature string(return type, method name and parameters */
    public String getSignatureString() {
        String res = "";
        sortRelatedObjects(new SortByOffset());
        String comma = "";
        for (Enumeration e = getModifiers();
             e.hasMoreElements();) {
            res += ((FrjModifier) e.nextElement()).getStr() + " ";
        }

        if (returnType != null) {
            res += returnType.getFullString() + " ";
        }

        res += this.getStr();
                
        res += "(";
        for (Enumeration e = getParameters();
             e.hasMoreElements();) {
            FrjVariable var = (FrjVariable) e.nextElement();
            res += comma + var.getType().getFullString() + " " + var.getStr();
            comma = ", ";
        }
        res += ")";
        return res;
    }
                
    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);

        traverseObjects(getModifiers(), tm);
        if (returnType != null) {
            returnType.traverse(tm);
        }

        tm.doMyself(this);

        traverseObjects(getParameters(), tm);
        traverseObjects(getThrowTypes(), tm);   
        if (body != null) {
            body.traverse(tm);
        }

        tm.doAfterTraverse(this);
    }

    /* static initializer */
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];

        sortNameTable[METHOD] = "Method";
        sortNameTable[CONSTRUCTOR] = "Constructor";
    }

    // tester   
    public static void main(String args[]) {
        if (args.length == 0) {
            System.err.println("java Japid.Freja.XXX ClassName");
            return;
        }
        FrjWorld.load(new FrejaEntry(FrjWorld.J_MODEL_PATH, args[0]));

        for (Enumeration e = FrjExtents.getObjects(CLASS_NAME);
             e.hasMoreElements();) {
            FrjRoutine r = (FrjRoutine) e.nextElement();
            FrjObject c
                = (FrjObject) r.getAncestorObject(FrjClass
                                                  .getClassNameStatically());
            FrjTypeList tl = r.parameter2TypeList();
            System.out.println(c.getStr() + "#" + r.getStr() + "(" + tl + ")");
        }
    }

}
