/* 
 * Program:             $RCSfile: FrjStatement.java,v $  $Revision: 41.5 $
 * 
 * Purpose:             The class corresponding to Statement.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjStatement.java,v 41.5 2003/02/12 03:39:38 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjStatement extends FrjObject {
    // Class variables
    /** A class name of this class */
    public final static String  CLASS_NAME = "Statement";

    /** table from a sort to string */
    protected static String sortNameTable[];    // initialized by static block
        
    // Attributes
    public int                  sort;

    // Related objects
    protected FrjObject         possessor = null;
    protected FrjStatement      parentStatement = null;
    protected Vector            childStatements = null;
    protected Vector            expressions = null;
    protected Vector            variables = null;
    protected Vector            labels = null;
    protected FrjClass          localClass = null;

    // Constatnts
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int        _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int     EMPTY = _i++;
    public final static int     EXPRESSION = _i++;
    /* public final static int   S_LABELED = _i++; */
    public final static int     BLOCK = _i++;
    public final static int     DEF_VAR = _i++;
    public final static int     IF = _i++;
    public final static int     SWITCH = _i++;
    public final static int     WHILE = _i++;
    public final static int     DO = _i++;
    public final static int     FOR = _i++;
    public final static int     BREAK = _i++;
    public final static int     CONTINUE = _i++;
    public final static int     RETURN = _i++;
    public final static int     THROW = _i++;
    public final static int     EXCEPTION = _i++;
    public final static int     TRY = _i++;
    public final static int     CATCH = _i++;
    public final static int     FINALLY = _i++;
    public final static int     SYNCHRONIZED = _i++;
    public final static int     SWITCH_LABELED_EMPTY = _i++;
    public final static int     DEF_LOCAL_CLASS = _i++;
    public final static int     ASSERT = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;         // sentinel


    // Constructors
    public FrjStatement(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        //this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }
                
    public FrjStatement(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }

    public FrjStatement() {
        super();
        sort = INVALID_ID;
    }


    // Class methods
    /** return the class name of this object statically */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }
                        
    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjStatement(entry, obj);
    }

    // Instance methods
    /** return the class name of this object */
    public String getClassName() {
        return CLASS_NAME;
    }

    // about attributes
    /** return the sort */
    public int getSort() {
        return sort;
    }
    /** check this object has a valid sort */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** return the sort as string */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return "INVALID SORT";
        }
        return sortNameTable[sort];
    }

    /** return the name of attribute and its value as string */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** convert object to string */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // about relations
    public void setRelatedObjects() {
        childStatements = new Vector();
        expressions = new Vector();
        variables = new Vector();
        labels = new Vector();

        super.setRelatedObjects();
    }
    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        possessor = null;
        parentStatement = null;
        childStatements = null;
        expressions = null;
        variables = null;
        labels = null;
        localClass = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.LOCAL_CLASS)) {
            localClass = (FrjClass) obj;
        } else if (relName.equals(FrjRelations.LOCAL_VARIABLE)) {
            variables.addElement(obj);
        } else if (relName.equals(FrjRelations.STATEMENT_LABEL)) {
            labels.addElement(obj);
        } else if (relName.equals(FrjRelations.STATEMENT_STATEMENT)) {
            if (rel.get1stObject().getID() == this.getID()) { // this is parent
                childStatements.addElement(obj);
            } else {      // this is child
                parentStatement = (FrjStatement) obj;
            }
        } else if (relName.equals(FrjRelations.STATEMENT_EXPRESSION)) {
            expressions.addElement(obj);
        } else if (relName.equals(FrjRelations.ROUTINE_BODY)
                 ||     relName.equals(FrjRelations.SBLOCK_BODY)) {
            possessor = obj;
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(childStatements);
        sorter.sort();
        sorter.setVector(expressions);
        sorter.sort();
        sorter.setVector(variables);
        sorter.sort();
    }

    /**
     * Traverses <I>statement_statement</I>.
     * @return the parent statement
     */
    public FrjStatement getParentStatement() {
        return parentStatement;
    }
    /**
     * Traverses <I>statement_statement</I>.
     * @return child statements
     */
    public Enumeration getChildStatements() {
        return childStatements.elements();
    }
    /** Traverses <I>statement_expression</I>. */
    public Enumeration getExpressions() {
        return expressions.elements();
    }
    /** Traverses <I>local_variable</I>. */
    public Enumeration getLocalVariables() {
        return variables.elements();
    }
    /** Traverses <I>statement_label</I>. */
    public Enumeration getLabels() {
        return labels.elements();
    }
    /**
     * Traverses <I>routine_body</I> or <I>sblock_body</I>.
     * @return an instance of FrjRoutine or FrjSblock
     */
    public FrjObject getPossessor() {
        return possessor;
    }

    /**
     * Traverses <I>local_class</I>.
     */
    public FrjClass getLocalClass() {
        return localClass;
    }

    /** return the parent object */
    public FrjObject getParentObject() {
        if (parentStatement != null) {
            return parentStatement;
        } else {
            return possessor;
        }
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjStatement) obj).sort = sort;
    }

    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);

        for (Enumeration e = getLabels();
             e.hasMoreElements();) {
            FrjLabel label = (FrjLabel) e.nextElement();
            label.traverse(tm);
        }

        tm.doMyself(this);

        traverseObjects(getChildStatements(), tm);
        traverseObjects(getLocalVariables(), tm);
        traverseObjects(getExpressions(), tm);
        if (localClass != null) {
            localClass.traverse(tm);
        }

        tm.doAfterTraverse(this);
    }

    /** return all child objects as Enumerations */
    /****
         public Enumeration getChildObjects() {
         Vector         objects = new Vector();

         for (Enumeration children = getChildStatements();
         children.hasMoreElements();) {
         objects.addElement(children.nextElement());
         }

         for (Enumeration exprs = getExpressions();
         exprs.hasMoreElements();) {
         objects.addElement(exprs.nextElement());
         }
                
         return objects.elements();
         }
    ****/


    /****
         public void addParentStatement(FrjStatement stmt) {
         FrjRelation rel
         = new FrjRelation(FrjRelations.STATEMENT_STATEMENT,
         stmt, this);
         }
    ****/

    /* static initializer */
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];

        sortNameTable[EMPTY] = "Empty";
        sortNameTable[EXPRESSION] = "Expression";
        sortNameTable[BLOCK] = "Block";
        sortNameTable[DEF_VAR] = "Define Variable";
        sortNameTable[IF] = "If";
        sortNameTable[SWITCH] = "Switch";
        sortNameTable[WHILE] = "While";
        sortNameTable[DO] = "Do";
        sortNameTable[FOR] = "For";
        sortNameTable[BREAK] = "Break";
        sortNameTable[CONTINUE] = "Continue";
        sortNameTable[RETURN] = "Return";
        sortNameTable[THROW] = "Throw";
        sortNameTable[EXCEPTION] = "Exception";
        sortNameTable[TRY] = "Try";
        sortNameTable[CATCH] = "Catch";
        sortNameTable[FINALLY] = "Finally";
        sortNameTable[SYNCHRONIZED] = "Synchronized";
        sortNameTable[SWITCH_LABELED_EMPTY] = "Switch Labeled Empty";
        sortNameTable[DEF_LOCAL_CLASS] = "Define Local Class";
        sortNameTable[ASSERT] = "Assert";
        //              sortNameTable[SORT_LAST_MEMBER] = null;         /* sentinel */
    }

    public static void main(String args[]) {
        System.out.println("First: " + SORT_FIRST_MEMBER);
        System.out.println("Empty: " + EMPTY);
        System.out.println("Sync: " + SYNCHRONIZED);
        System.out.println("Last: " + SORT_LAST_MEMBER);
    }
}
