/* 
 * Program:             $RCSfile: FrjType.java,v $  $Revision: 41.20 $
 * 
 * Purpose:             The class corresponding to type.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjType.java,v 41.20 2003/02/15 11:56:30 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjType extends FrjObject {
    // Class variables.
    /** A class name of this class. */
    public final static String  CLASS_NAME = "Type";

    /** Table from a sort to string. */
    protected static String sortNameTable[];    // Initialized by static block.

    // Attributes.
    public int                          sort;
    public String                       pkgName;

    // Related objects.
    /** NOTE: if A[], [] is a parent of A */
    protected FrjType           parentType = null;
    protected FrjType           childType = null;
    protected FrjQName          qname = null;
    protected FrjExpression     referrer = null;
    protected FrjClassInterface referringClassInterface = null;
    protected Vector            possessors = null;

    // Constatnts.
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected final static int  PKGNAME_INDEX = SORT_INDEX +1;
    protected static int        _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i;         // Sentinel.
    public final static int     PRIMITIVE = _i++;
    public final static int     OBJECT = _i++;
    public final static int     ARRAY = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // Sentinel.

    public final static String VOID = "void",
        BYTE = "byte", CHAR = "char",
        INT = "int", SHORT = "short", LONG = "long",
        FLOAT = "float", DOUBLE = "double", BOOLEAN = "boolean";

    private final static Vector typeOrder = new Vector();
    // Initialized in SBLock.

    public FrjType(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        // this.sort = Integer.parseInt(attrs[SORT_INDEX]);
        // this.pkgName = attrs[PKGNAME_INDEX];
    }
    public FrjType(String str) {
        super();
        setStr(str);
        relations = new FrjRelations();
        // FrjObject() doesn't add to Extents.
        FrjExtents.addObject(this);
    }

    public FrjType(FrejaEntry entry, String str) {
        this(entry, str, null);
    }

    public FrjType(FrejaEntry entry, String str, String pkgName) {
        super(entry);
        setStr(str);
        if (sort == OBJECT) {
            this.pkgName = pkgName;
        }
        relations = new FrjRelations();
    }

    public FrjType() {
        super();
        sort = INVALID_ID;
    }

    protected void setStr(String str) {
        if (str.equals(VOID) || str.equals(BYTE) || str.equals(CHAR)
            || str.equals(INT) || str.equals(SHORT) || str.equals(LONG)
            || str.equals(FLOAT) || str.equals(DOUBLE)
            || str.equals(BOOLEAN)) {
            this.str = str;
            sort = PRIMITIVE;
        } else if (str.equals("[]")) {
            this.str = str;
            sort = ARRAY;
        } else {
            this.str = str;
            sort = OBJECT;
        }
    }

    protected void setPackageName(FrejaEntry entry) {
        if (sort == OBJECT) {
            pkgName 
                = referringClassInterface.getFrejaEntry().getPackageName();
        }
    }

    public String getPackageName() {
        return pkgName; 
    }

    // Class methods
    /** return the class name of this object statically */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjType(entry, obj);
    }

    // Instance methods
    /** return the class name of this object */
    public String getClassName() {
        return CLASS_NAME;
    }

    // For attribute.
    /** return the sort */
    public int getSort() {
        return sort;
    }

    /** check this object has a valid sort */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }

    /** return the sort as string */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return "INVALID SORT";
        }
        return sortNameTable[sort];
    }

    /** return the name of attribute and its value as string */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        attrs[PKGNAME_INDEX] = new FrjAttrsAsString("pkgName",
                                                    pkgName);
        return attrs;
    }

    /**
     * Returns full string.
     * Full string means int[], java.util.Vector and so on.
     * 
     * NOTE: If type appears Vector in source code, returns only "Vector",
     *       and if java.util.Vector, returns "java.util.Vector".
     */
    public String getFullString() {
        String res = "";
        if (qname != null) {
            res += qname.getFullString() + "." ;
        }
        if (childType != null) {
            res += childType.getFullString();
        }
        res += this.getStr();
        return res;
    }

    /** convert object to string */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        res += ", package: " + pkgName;
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        res += stringToSdbformat(pkgName) + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    /** @deprecated API. Use isNullType(). */
    // Commented by S.Yamamoto 2002/09/10.
    // public boolean isNullObjectiveType() {
    //     return (sort == OBJECT && (str == null || str.equals("null")));
    // }

    public boolean isNullType() {
        return ((sort == OBJECT || sort == ARRAY) &&
                (str == null || str.equals("null")));
    }

    // For relations.
    public void setRelatedObjects() {
        possessors = new Vector();
        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        parentType = null;
        childType = null;
        qname = null;
        referrer = null;
        referringClassInterface = null;
        possessors = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {
        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.TYPE_TYPE)) {
            if (rel.get1stObject().getID() == this.getID()) {
                childType = (FrjType)obj;
            } else {
                parentType = (FrjType)obj;
            }
        } else if (relName.equals(FrjRelations.TYPE_QNAME)) {
            qname = (FrjQName)obj;
        } else if (relName.equals(FrjRelations.REFER_TYPE)) {
            referrer = (FrjExpression)obj;
        } else if (relName.equals(FrjRelations.TYPE_CLASS)) {
            referringClassInterface = (FrjClass)obj;
            setPackageName(referringClassInterface.getFrejaEntry());
        } else if (relName.equals(FrjRelations.TYPE_INTERFACE)) {
            referringClassInterface = (FrjInterface)obj;
            setPackageName(referringClassInterface.getFrejaEntry());
        } else if (relName.equals(FrjRelations.EXTEND_CLASS)
                   || relName.equals(FrjRelations.IMPLEMENT_INTERFACE)
                   || relName.equals(FrjRelations.EXTEND_INTERFACE)
                   || relName.equals(FrjRelations.RETURN_TYPE)
                   || relName.equals(FrjRelations.THROW_TYPE)
                   || relName.equals(FrjRelations.VARIABLE_TYPE)
                   || relName.equals(FrjRelations.HAS_TYPE)) {
            possessors.addElement(obj);
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }

    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(possessors);
        sorter.sort();
    }
        
    /**
     * Traverses <I>type_type</I>.
     * @return the parent type<P>
     * example:<P>
     * <PRE>   String []  </PRE> parent: [], child: String
     * <PRE>   String [][]</PRE> parent: [], child: String[]
     */
    public FrjType getParentType() {
        return parentType;
    }

    /**
     * Traverses <I>type_type</I>.
     * @return the child type<P>
     */
    public FrjType getChildType() {
        return childType;
    }

    /** Traverses <I>type_qname</I>. */
    public FrjQName getQName() {
        return qname;
    }

    /**
     * Traverses <I>type_class</I> or <I>type_interface</I>.
     * @return FrjClass that this object referes.
     * If auto loading other classes, searches Creja for referring class.
     * @see Japid.Freja.FrjWorld#setAutoLoading()
     * @see Japid.Freja.FrjWorld#isAutoLoading()
     */
    public FrjClassInterface getReferringClassInterface() {
        if (sort == OBJECT && referringClassInterface == null) {
            if (isNullType()) {
                return null;
            }

            Enumeration fs = FrjExtents.getObjects(getFrejaEntry(),
                                                   FrjFile.getClassNameStatically());
            FrjFile f = (FrjFile) fs.nextElement();
            if (f == null) {
                FrjWorld.warning("Type " + this + " has no parent.");
                return null;
            }
            String className;
            if (pkgName != null && !pkgName.equals("")) {
                className = pkgName + "." + str;
            } else if (qname == null) {
                className = str;
            } else {
                className = qname.getFullString() + "." + str;
            }
            FrejaEntry entry
                = Creja.getFrejaEntry(f.getDefinedPackage(),
                                      f.getImportedPackages(), className);
            if (entry != null) {
                if (FrjExtents.isLoaded(entry)) {
                    entry = FrjExtents.getLoadedFrejaEntry(entry);
                } else {
                    if (FrjWorld.isAutoLoading()) {
                        FrjWorld.load(entry);
                    } else {
                        return null;
                    }
                }
                // Search class.
                for (Enumeration e = FrjExtents
                         .getObjects(entry.getFrejaID(),
                                     FrjClass.getClassNameStatically());
                     e.hasMoreElements();) {
                    FrjClass c = (FrjClass) e.nextElement();
                    if (c.getStr().equals(str)) {
                        new FrjRelation(FrjRelations.TYPE_CLASS, this, c);
                        resetRelatedObjects();
                        c.resetRelatedObjects();
                        // Just for debug.
                        if (referringClassInterface == null) {
                            FrjWorld.vwarning("Invalid type_class.\n" +
                                              this + "\n  " + c);
                        }
                        return referringClassInterface;
                    }
                }

                // Search interface.
                for (Enumeration e = FrjExtents
                         .getObjects(entry.getFrejaID(),
                                     FrjInterface.getClassNameStatically());
                     e.hasMoreElements();) {
                    FrjInterface i = (FrjInterface)e.nextElement();
                    if (i.getStr().equals(str)) {
                        new FrjRelation(FrjRelations.TYPE_INTERFACE, this, i);
                        resetRelatedObjects();
                        i.resetRelatedObjects();
                        // Just for debug.
                        if (referringClassInterface == null) {
                            FrjWorld.vwarning("Invalid type_interface.\n" +
                                                    this + "\n  " + i);
                        }
                        return referringClassInterface;
                    }
                }
            }

            // A referringClassInterface is not searched.
            String      pkgs = "";
            FrjQName    pkg;

            pkg = f.getDefinedPackage();
            if (pkg != null) {
                pkgs  += pkg.getFullString() + " ";
            }
            for (Enumeration p = f.getImportedPackages();
                 p.hasMoreElements();) {
                pkgs += ((FrjQName)p.nextElement()).getFullString() + " ";
            }
            FrjWorld.vwarning("Undefined Class \"" + className +
                                    "\" in \"" + f.getStr() + "\".\n" +
                                    "\t" + this + "\n" +
                                    "\tpackages: \"" + pkgs + "\".");
        }
        return referringClassInterface;
    }

    /** Traverses <I>refer_type</I>. */
    public FrjExpression getReferrer() {
        return referrer;
    }
    /**
     * Traverses <I>extend_class</I> or <I>implement_interface</I>
     * or <I>extend_interface</I> or <I>return_type</I> or <I>throw_type</I>
     * or <I>variable_type</I> or <I>has_type</I>.
     */
    public Enumeration getPossessors() {
        return possessors.elements();
    }

    public FrjObject getParentObject() {
        if (parentType != null) {
            return parentType;
        } else if (referrer != null) {
            return referrer;
        } else {
            FrjObject parent = null;
            for (Enumeration e = getPossessors();
                 e.hasMoreElements();) {
                parent = (FrjObject) e.nextElement();
                if (getFrejaEntry() != parent.getFrejaEntry()) {
                    parent = null;
                    continue;
                }
                if (parent instanceof FrjRoutine
                    || parent instanceof FrjVariable
                    || parent instanceof FrjClassInterface) {
                    // return or throw type of method, or type of variable,
                    // or class of extended,
                    // or inteface of extended or implemented
                    break;
                }
            }
            if (parent == null) {
                if (referringClassInterface != null &&
                    getFrejaEntry() == referringClassInterface.getFrejaEntry()) {
                    parent = referringClassInterface;
                } else {
                    Enumeration files = FrjExtents.getObjects(getFrejaEntry(),
                                                              FrjFile.getClassNameStatically());
                    parent = (FrjObject) files.nextElement();
                }
            }
            return parent;
        }
        // Dead code.
        // return null;
    }

    /** Sets referringClassInterface of all objective types. */
    public static void setReferringClassInterfaces() {
        setReferringClassInterfaces(FrjExtents.getObjects(FrjType.CLASS_NAME));
    }

    public static void setReferringClassInterfaces(FrejaEntry entry) {
        setReferringClassInterfaces(FrjExtents.getObjects(entry,
                                                          FrjType.CLASS_NAME));
    }

    protected static void setReferringClassInterfaces(Enumeration e) {
        while (e.hasMoreElements()) {
            FrjType t = (FrjType) e.nextElement();
            t.getReferringClassInterface();
        }
    }
                
    // Equality.
    /**
     * If parameter's type equals to this type, return true.
     *   e.g. int equals to int
     *        double doesn't equal to float
     *        int [] doesn't equal to int
     *        Vector probablly equals to Vector
     *        Vector probablly equals to java.util.Vector
     *          (You may define your original class 'My.util.Vector.')
     */
    public boolean equals(FrjType type) {
        if (this.isSame(type)) {
            return true;
        }

        if (this.getSort() != type.getSort()) {
            return false;
        }
                
        // Same sort.
        if (this.getSort() == PRIMITIVE) {
            return (this.getStr()).equals(type.getStr());
        } else if (this.getSort() == ARRAY) {
            if (type.getSort() == ARRAY) {
                return (this.childType).equals(type.childType);
            } else {
                return false;
            }
        } else if (this.getSort() == OBJECT) {
            /*** must consider package environment ***/
            if (type.getSort() == OBJECT) {
                if (this.getStr() == "null" || type.getStr() == "null") {
                    return true;
                }
                try {
                    return (this.getStr()).equals(type.getStr());
                } catch (Exception e) {
                    System.err.print("this: " + this);  
                    System.err.println((this.getStr() == null) ? " <NULL>": "");
                    System.err.println("type: " + type);
                    System.err.println((type.getStr() == null) ? " <NULL>": "");
                    System.exit(1);
                }
            } else {
                return false;
            }
        }
        return false;
    }

    /**
     * If parameter's type is compatible to this type, return true.
     *   e.g. int is compatible to int
     *        float is compatible to double
     *        double is NOT compatible to float
     *        int [] is not compatible to int
     *        Vector is compatible to Object
     *        Object is NOT compatible to Vector
     *          (You must pay attention to order.)
     */
    public boolean isCompatible(FrjType type) {
        if (this.equals(type)) {
            return true;
        }

        if (this.getStr() == "String" || type.getStr() == "String") {
            // String is compatible for any type.
            return true;
        }
        // "null" is compatible for any object and array.
        if (this.isNullType() && 
            (type.getSort() == OBJECT || type.getSort() == ARRAY)) {
            return true;
        }
        // Object is compatible for any object and array
        if ((this.getSort() == OBJECT || this.getSort() == ARRAY) &&
            type.getStr().equals("Object")) {
            return true;
        }

        if (this.getSort() != type.getSort()) {
            return false;
        }
                
        // Same sort.
        if (this.getSort() == PRIMITIVE) {
            int thisOrder = typeOrder.indexOf(this.getStr());
            int argOrder = typeOrder.indexOf(type.getStr());
            if (thisOrder < 0 || argOrder < 0) {  // Void or boolean.
                return false;
            }
            return (thisOrder <= argOrder);
        } else if (this.getSort() == ARRAY) {
            return (this.childType).isCompatible(type.childType);
        } else if (this.getSort() == OBJECT) {
            if (this.getStr().equals(type.getStr())) {
                return true;
            } else if (type.getStr().equals("Object")) {
                // Object is the parent of all class.
                return true;
            } else {
                if (FrjWorld.isAutoLoading()) {
                    FrjClassInterface c = this.getReferringClassInterface();
                    if (c == null) {
                        FrjWorld.vwarning("FrjType#isCompatible(): " +
                            "getReferringClassInterface() returns null.");
                    } else { 
                        for (Enumeration e = c.getSuperClassInterfaceTypes();
                             e.hasMoreElements();) {
                            FrjType superType = (FrjType)e.nextElement();
                            if (superType.isCompatible(type)) {
                                return true;
                            }
                        }
                    }
                }
            }
        }
        return false;
    }
                
    /**
     * Type conversion.
     * e.g. int and char -> int
     *      long and int -> int
     */
    public FrjType convert(FrjType type) throws TypesNotConvertableException {
        if (type == null) {     // e.g. this.a 
            return this;
        } else if (getStr().equals("String")) {
            // String is convertible any type
            // NOTE: must check whethear String's sub type
            return this;
        } else if (type.getStr().equals("String")) {
            return type;
        } else if (getSort() == PRIMITIVE && type.getSort() == PRIMITIVE) {
            int type1 = typeOrder.indexOf(this.getStr());
            int type2 = typeOrder.indexOf(type.getStr());
            if (type1 < 0 || type2 < 0) {  // void or boolean
                throw new TypesNotConvertableException(this, type);
            } else if (type1 < type2) {
                return type;
            } else {
                return this;
            }
        } else {
            throw new TypesNotConvertableException(this, type);
        }
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjType) obj).sort = sort;
    }

    // Traverse method.
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);

        if (qname != null) {
            qname.traverse(tm);
        }
        if (childType != null) {
            childType.traverse(tm);
        }

        tm.doMyself(this);
        tm.doAfterTraverse(this);
    }
                
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];
                
        sortNameTable[PRIMITIVE] = "Primitive Type";
        sortNameTable[OBJECT] = "Objective Type";
        sortNameTable[ARRAY] = "Array Type";

        typeOrder.addElement(BYTE);
        typeOrder.addElement(CHAR);
        typeOrder.addElement(INT);
        typeOrder.addElement(SHORT);
        typeOrder.addElement(LONG);
        typeOrder.addElement(FLOAT);
        typeOrder.addElement(DOUBLE);
    }

    // Tester.
    public static void main(String args[]) {
        if (args.length == 0) {
            System.err.println("java Japid.Freja.XXX ClassName");
            return;
        }
        FrjWorld.load(new FrejaEntry(FrjWorld.J_MODEL_PATH, args[0]));
        for (Enumeration e = FrjExtents.getObjects(CLASS_NAME);
             e.hasMoreElements();) {
            FrjType obj = (FrjType) e.nextElement();
            obj.setRelatedObjects();
        }
        FrjType intType = new FrjType();
        intType.sort = PRIMITIVE;
        intType.str = "int";

        FrjType stringType = new FrjType();
        stringType.sort = OBJECT;
        stringType.str = "String";

        FrjType arrayType = new FrjType();
        arrayType.sort = ARRAY;
        arrayType.str = "[]";
        arrayType.childType = stringType;
                
        for (Enumeration e = FrjExtents.getObjects(CLASS_NAME);
             e.hasMoreElements();) {
            FrjType type = (FrjType) e.nextElement();
            if (type.parentType == null) {
                System.out.print(type.getStr());
                for (FrjType obj = type.childType;
                     obj != null; obj = obj.childType) {
                    System.out.print("->" + obj.getStr());
                }
            }
            if (type.equals(intType)) {
                System.out.print(" is equals to int.");
            } else if (type.equals(stringType)) {
                System.out.print(" is equals to String.");
            } else if (type.equals(arrayType)) {
                System.out.print(" is equals to String[].");
            }
            System.out.println();
        }
    }
}
