/* 
 * Program:             $RCSfile: FrjVariable.java,v $  $Revision: 41.4 $
 * 
 * Purpose:             The class corresponding to Expression.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: FrjVariable.java,v 41.4 2003/02/12 03:39:38 yamamoto Exp yamamoto $
 */

package Japid.Freja;

import java.io.*;
import java.util.*;
import Japid.Freja.DBAccess.*;

public class FrjVariable extends FrjIdentifier {
    // Class variables
    /** A class name of this class */
    public final static String  CLASS_NAME = "Variable";

    /** table from a sort to string */
    protected static String     sortNameTable[];        // initialized by static block

    // Attribute
    public int                          sort;

    // Related objects
    protected FrjObject         definer;        // an object defining this 
    protected FrjType           type;
    protected Vector            referrers;      // objects referring this
    protected FrjExpression     initializer;
    //  protected Vector                modifiers = null;

    // Constatnts
    protected final static int  SORT_INDEX = ORIGINAL_ATTR_START;
    protected static int        _i = 0;
    protected final static int  SORT_FIRST_MEMBER = _i; // sentinel
    public final static int             CLASS_MEM = _i++;
    public final static int             INTERFACE_MEM = _i++;
    public final static int             PARAMETER = _i++;
    public final static int             LOCAL_VAR = _i++;
    protected final static int  SORT_LAST_MEMBER = --_i;        // sentinel
                
    // Constructors
    public FrjVariable(FrejaEntry entry, DBClassObject obj) {
        super(entry, obj);
        //this.sort = Integer.parseInt(attrs[SORT_INDEX]);
    }

    public FrjVariable(FrejaEntry entry) {
        super(entry);
        sort = INVALID_ID;
    }

    public FrjVariable() {
        super();
        sort = INVALID_ID;
    }


    // Class methods
    /** return the class name of this object statically */
    public static String getClassNameStatically() {
        return CLASS_NAME;
    }

    /** create object from SDB */
    protected void create(FrejaEntry entry, DBClassObject obj) {
        new FrjVariable(entry, obj);
    }


    // Instance methods
    /** return the class name of this object */
    public String getClassName() {
        return CLASS_NAME;
    }

    // about attribute
    /** return the sort */
    public int getSort() {
        return sort;
    }
    /** check this object has a valid sort */
    public boolean hasValidSort() {
        return (SORT_FIRST_MEMBER <= sort && sort <= SORT_LAST_MEMBER);
    }
    /** return the sort as string */
    public String getSortString() {
        if (!hasValidSort()) {  /* error */
            return null;
        }
        return sortNameTable[sort];
    }

    /** return the name of attribute and its value as string */
    public FrjAttrsAsString[] getAttrsAsString() {
        FrjAttrsAsString attrs[] = super.getAttrsAsString();
                
        attrs[SORT_INDEX] = new FrjAttrsAsString("sort", getSortString());
        return attrs;
    }

    /** convert object to string */
    public String toString() {
        String res;
        res = super.toString();
        res += ", sort: " + getSortString();
        return res;
    }

    protected String toSdbline() {
        String res = super.toSdbline();
        res += getSort() + JDBConsts.SDB_ATTR_SEPARATOR;
        return res;
    }

    // about relation
    public void setRelatedObjects() {
        modifiers = new Vector();
        referrers = new Vector();

        super.setRelatedObjects();
    }

    public void removeRelatedObjects() {
        super.removeRelatedObjects();
        definer = null;
        type = null;
        referrers = null;
        initializer = null;
        modifiers = null;
    }

    protected void setRelatedObject(FrjObject obj, FrjRelation rel)
        throws NoSuchRelatedObjectException {

        String relName = rel.getRelationName();
        if (relName.equals(FrjRelations.VARIABLE_TYPE)) {
            type = (FrjType) obj;
        } else if (relName.equals(FrjRelations.VARIABLE_MODIFIER)) {
            modifiers.addElement((FrjModifier) obj);
        } else if (relName.equals(FrjRelations.INITIALIZE_VARIABLE)) {
            initializer = (FrjExpression) obj;
        } else if (relName.equals(FrjRelations.CLASS_VARIABLE)
                 || relName.equals(FrjRelations.INTERFACE_VARIABLE)
                 || relName.equals(FrjRelations.PARAMETER_VARIABLE)
                 || relName.equals(FrjRelations.LOCAL_VARIABLE)) {
            if (definer != null) {      
                String msg = "Variable " + this + " has already defined.\n";
                msg += "old: " + definer + " new: " + obj;
                throw new NoSuchRelatedObjectException(msg);
            }
            definer = obj;
        } else if (relName.equals(FrjRelations.REFER_VARIABLE)) {
            referrers.addElement(obj);
        } else if (relName.equals(FrjRelations.SYMBOL_COMMENT)) {
            setRelatedComment(obj);
        } else {
            throw new NoSuchRelatedObjectException(rel);
        }
    }
    public void sortRelatedObjects(VectorSorter sorter) {
        sorter.setVector(modifiers);
        sorter.sort();
        sorter.setVector(referrers);
        sorter.sort();
    }

    /**
     * Traverses <I>class_variable</I> or <I>interface_variable</I>
     * or <I>parameter_variable</I> or <I>local_variable<I>.
     *
     * @return FrjObject which defines this variable
     */
    public FrjObject getDefiner() {
        return definer;
    }

    /** Traverses <I>variable_type</I>. */
    public FrjType getType() {
        return type;
    }
    /** Traverses <I>refer_variable</I>. */
    public Enumeration getReferrers() {
        return referrers.elements();
    }
    /** Traverses <I>variable_initializer</I>. */
    public FrjExpression getInitializer() {
        return initializer;
    }

    public FrjObject getParentObject() {
        return definer;
    }

    public void copyTo(FrjObject obj) {
        super.copyTo(obj);
        ((FrjVariable) obj).sort = sort;
    }

    // traverse method
    public void traverse(TraverseMethod tm) {
        tm.doBeforeTraverse(this);

        traverseObjects(getModifiers(), tm);
        type.traverse(tm);

        tm.doMyself(this);

        if (initializer != null) {
            initializer.traverse(tm);
        }

        tm.doAfterTraverse(this);
    }
        
    // static initializer
    static {
        sortNameTable = new String[SORT_LAST_MEMBER +1];
                
        sortNameTable[CLASS_MEM] = "Class Member";
        sortNameTable[INTERFACE_MEM] = "Interface Member";
        sortNameTable[PARAMETER] = "Parameter";
        sortNameTable[LOCAL_VAR] = "Local variable";
    }
}
