/* 
 * Program:             $RCSfile: ModifiableStringBuffer.java,v $  $Revision: 41.12 $
 * 
 * Purpose:             Modifiale string buffer.
 *
 * Author:              Y.Hachisu  1998/06/18
 *                      S.Yamamoto  1998/06/18
 *
 * (C) Copyright:       Y.Hachisu and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: ModifiableStringBuffer.java,v 41.12 2003/03/27 09:10:31 yamamoto Exp yamamoto $
 */

package Japid.Freja;

public class ModifiableStringBuffer {
    public final static int     UNKNOWN = -1, LF = 0, CR = 1, CRLF = 2, LFCR = 3;
    protected byte              buffer[];
    protected int               strLen = 0;
    // protected static int        BUFF_SIZE = 256;
    protected static int        BUFF_SIZE = 16 * 1024;

    private int                 lineEndType = UNKNOWN;
    protected String            encoding = null;

    public ModifiableStringBuffer() {
        buffer = new byte[BUFF_SIZE];
    }

    public ModifiableStringBuffer(int size) {
        if (size < 0) {
            throw new NegativeArraySizeException();
        }
        buffer = new byte[size];
    }

    public ModifiableStringBuffer(String str) {
        buffer = str2bytes(str);
        strLen = buffer.length;
    }

    /**
     * sets encoding of String
     */
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }

    /**
     * Appends the string to this modifiable string.
     * @param str - a string
     * Returns this modifiable string buffer
     */
    public ModifiableStringBuffer append(String str) {
        byte strBytes[] = str2bytes(str);
        int len = strBytes.length;
        if (buffer.length - strLen <= len) {
            expand(buffer.length + len);
        }
        System.arraycopy(buffer, strLen, strBytes, 0, len);
        strLen += len;
        return this;
    }

    /**
     * Appends the byte to this modifiable string.
     * @param b - array of byte
     * @param len - length of array
     * Returns this modifiable string buffer
     */
    public ModifiableStringBuffer append(byte b[], int len) {
        if (buffer.length - strLen <= len) {
            expand(buffer.length + len);
        }
        System.arraycopy(b, 0, buffer, strLen, len);
        strLen += len;
        return this;
    }


    /**
     * Deletes n-characters from offset.
     * @param offset - offset of deleting characters
     * @param length - length of deleting characters (means n-)
     * @returns this modifialbe string buffer
     */
    public ModifiableStringBuffer delete(int offset, int length) {
        replace(offset, length, "");
        return this;
    }

    /**
     * Inserts characters from offset.
     * @param offset - offset of inserting characters
     * @param str - inserting string 
     * @returns this modifialbe string buffer
     */
    public ModifiableStringBuffer insert(int offset, String str) {
        replace(offset, 0, str);
        return this;
    }

    /**
     * Replace n-characters from offset with str. 
     * @param offset - offset of inserting characters
     * @param length - length of deleting characters
     * @param str - replacing string 
     * @returns this modifialbe string buffer
     */
    public ModifiableStringBuffer replace(int offset, int length, String str) {
        byte strBytes[] = str2bytes(str);
        if (offset < 0) {
            throw new StringIndexOutOfBoundsException(offset);
        }
        if (offset + length > strLen) {
            throw new StringIndexOutOfBoundsException(offset+length);
        }
        if (length < 0) {
            throw new StringIndexOutOfBoundsException("Length must be zero or positive number.");
        }
        if (str == null) {
            throw new NullPointerException("Replacing strings must not be null.");
        }
        byte newBuffer[] = new byte[buffer.length + strBytes.length];
        int buffi;
        for (buffi = 0; buffi < offset; buffi++) {
            newBuffer[buffi] = buffer[buffi];
        }
        for (int i = 0; i < strBytes.length; i++, buffi++) {
            newBuffer[buffi] = strBytes[i];
        }
        for (int i = offset + length; i < strLen; i++, buffi++) {
            newBuffer[buffi] = buffer[i];
        }
        buffer = newBuffer;
        strLen = buffi;
        return this;
    }

    /**
     * @returns the numberes of characters in this buffer.<P>
     */
    public int length() {
        return strLen;
    }

    /**
     * @returns current capacity of this buffer.<P>
     * 
     * public int capacity() {
     *     return strLen;
     * }
     */

    /**
     * @param beginIndex - the beginning index, inclusive
     * @param endIndex - the ending index, exclusive
     * @returns the substring from beginIndex to endIndex<P>
     */
    public String substring(int begin, int end) {
        /*
          char resultBuff[] = new char[end - begin];
          for (int i = begin, j = 0; i < end; i++, j++) {
          resultBuff[j] = buffer[i];
          }
          return String.valueOf(resultBuff);
        */
        if (encoding == null) {
            return new String(buffer, begin, end - begin);
        } else {
            try {
                return new String(buffer, begin, end - begin, encoding);
            } catch (java.io.UnsupportedEncodingException e) {
                FrjWorld.warning("Unknown encoding: " + encoding);
                return null;
            }
        }
    }

    /**
     * @param beginIndex - the beginning index, inclusive
     * @returns the substring from beginIndex to end of this string<P>
     */
    public String substring(int begin) {
        return substring(begin, strLen);
    }

    /**
     * Returns the character at the specified index. An index ranges
     * from <tt>0</tt> to <tt>length() - 1</tt>.
     * @param index     the index of the desired character
     * @exception       StringIndexOutOfBoundsException If the index is not
     *                  in the range <tt>0</tt> to <tt>length()-1</tt>.
     */
    public char charAt(int index) {
        if (index < 0 || strLen <= index) {
            throw new StringIndexOutOfBoundsException(index);                   
        }
        return (char) buffer[index];
    }

    public String toString() {
        return substring(0, strLen);
    }

    /**
     * @param begin - the beginning index, inclusive
     * @param end - the ending index, exclusive
     * @returns the bytes from beginIndex to endIndex<P>
     */
    public byte[] getBytes(int begin, int end) {
        int len = end - begin;
        byte b[] = new byte[len];
        for (int i = 0; i < len; i++) {
            b[i] = buffer[begin + i];
        }
        return b;
    }

    /**
     * @returns the bytes
     */
    public byte[] getBytes() {
        return getBytes(0, strLen);
    }

    private void expand() {
        expand(buffer.length * 2);
    }

    private void expand(int size) {
        byte newBuffer[] = new byte[size];
        System.arraycopy(buffer, 0, newBuffer, 0, buffer.length);
        /*
         * for (int i = 0; i < buffer.length; i++) {
         *     newBuffer[i] = buffer[i];
         * }
         */
        buffer = newBuffer;
    }

    /** 
     *  return line-end type of this String
     *    a type is one of UKNOWN, LF, CR, CRLF, LFCR
     */
    public int getLineEndType() {
        if (lineEndType == UNKNOWN) {
            checkLineEndType();
        }
        return lineEndType;
    }

    private void checkLineEndType() {
        int leTypeCnt[] = new int[LFCR + 1];

        /* Count number of each line-end code. */
        for (int i = 0; i < strLen; i++) {
            if ((char)buffer[i] == '\n') {
                if ((i + 1 < strLen) && (char)buffer[i+1] == '\r') {
                    leTypeCnt[LFCR]++;
                    i++;
                } else {
                    leTypeCnt[LF]++;
                }
            } else if ((char)buffer[i] == '\r') {
                if ((i + 1 < strLen) && (char)buffer[i+1] == '\n') {
                    leTypeCnt[CRLF]++;
                    i++;
                } else {
                    leTypeCnt[CR]++;
                }
            }
        }

        /* Check max count of line-end code. */
        lineEndType = LF;
        for (int i = LF + 1; i <= LFCR; i++) {
            if (leTypeCnt[i] > leTypeCnt[lineEndType]) {
                lineEndType = i;
            }
        }

        if (leTypeCnt[lineEndType] == 0) { /* No line-end. */
            lineEndType = UNKNOWN;
        }
    }

    private byte[] str2bytes(String str) {
        byte b[] = null;
        if (encoding == null) {
            b = str.getBytes();
        } else {
            try {
                b = str.getBytes(encoding);
            } catch (java.io.UnsupportedEncodingException e) {
                FrjWorld.warning("Unknown encoding: " + encoding);
            }
        }
        return b;
    }

    public static void main(String args[]) {
        ModifiableStringBuffer str = new ModifiableStringBuffer();
        str.append(args[0]);
        System.out.print(str);
        str.replace(Integer.valueOf(args[1]).intValue(),
                    Integer.valueOf(args[2]).intValue(),
                    args[3]);
        System.out.println(" -> " + str);
    }
}
