/* 
 * Program:             $RCSfile: ByteBuffer.java,v $  $Revision: 40.5 $
 * 
 * Purpose:             
 *
 * Author:              K.Hoshino  2000/07/08
 *                      S.Yamamoto  2000/07/08
 *
 * (C) Copyright:       K.Hoshino and S.Yamamoto  2000 - 2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: ByteBuffer.java,v 40.5 2003/02/15 11:56:46 yamamoto Exp yamamoto $
 */

/*
 * See "java.lang.StringBuffer".
 */

package Japid.Tools.jspie;

public final class ByteBuffer /* implements java.io.Serializable */{
    
    private byte value[];
    
    private int count;
    
    private boolean shared;
    
    /* static final long serialVersionUID = 3388685877147921107L; */
    
    public ByteBuffer() {
        this(16);
    }
    
    public ByteBuffer(int length) {
        value = new byte[length];
        shared = false;
    }
    
    public ByteBuffer(String str) {
        this(str.length() * 2 + 16);
        append(str);
    }
    
    public int length() {
        return count;
    }
    
    public int capacity() {
        return value.length;
    }
    
    private final void copy() {
        byte newValue[] = new byte[value.length];
        System.arraycopy(value, 0, newValue, 0, count);
        value = newValue;
        shared = false;
    }
    
    public synchronized void ensureCapacity(int minimumCapacity) {
        if (minimumCapacity > value.length) {
            expandCapacity(minimumCapacity);
        }
    }
    
    private void expandCapacity(int minimumCapacity) {
        int newCapacity = (value.length + 1) * 2;
        if (minimumCapacity > newCapacity) {
            newCapacity = minimumCapacity;
        }
        
        byte newValue[] = new byte[newCapacity];
        System.arraycopy(value, 0, newValue, 0, count);
        value = newValue;
        shared = false;
    }
    
    public synchronized void setLength(int newLength) {
        if (newLength < 0) {
            throw new StringIndexOutOfBoundsException(newLength);
        }
        
        if (newLength > value.length) {
            expandCapacity(newLength);
        }
        
        if (count < newLength) {
            if (shared) copy();
            for (; count < newLength; count++) {
                value[count] = (byte) '\0';
            }
        } else {
            count = newLength;
            if (shared) copy();
        }
    }
    
    public synchronized byte charAt(int index) {
        if ((index < 0) || (index >= count)) {
             throw new StringIndexOutOfBoundsException(index);
        }
        return value[index];
    }
    
    public synchronized void getChars(int srcBegin, int srcEnd, byte dst[], int dstBegin) {
        if ((srcBegin < 0) || (srcBegin >= count)) {
            throw new StringIndexOutOfBoundsException(srcBegin);
        }
        if ((srcEnd < 0) || (srcEnd > count)) {
            throw new StringIndexOutOfBoundsException(srcEnd);
        }
        if (srcBegin < srcEnd) {
            System.arraycopy(value, srcBegin, dst, dstBegin, srcEnd - srcBegin);
        } else {
            if (srcBegin > srcEnd) {
                throw new StringIndexOutOfBoundsException
                    ("ByteBuffer.getChars(): begin > end");
            }
            /* We do nothing when srcBegin == srcEnd. */
        }
    }
    
    public synchronized void setCharAt(int index, char ch) {
        if ((index < 0) || (index >= count)) {
            throw new StringIndexOutOfBoundsException(index);
        }
        if (shared) copy();
        
        byte ch_byte[] = (String.valueOf(ch)).getBytes();
        
        insert(index, ch_byte, 0, ch_byte.length);      
    }
    
    public synchronized void setCharAt(int index, byte ch) {
        if ((index < 0) || (index >= count)) {
            throw new StringIndexOutOfBoundsException(index);
        }
        if (shared) copy();
        value[index] = ch;
    }
    
    public synchronized ByteBuffer append(Object obj) {
        return append(String.valueOf(obj));
    }
    
    public synchronized ByteBuffer append(String str) {
        if (str == null) {
            str = String.valueOf(str);
        }
        
        byte str_byte[] = str.getBytes();
        
        return append(str_byte);
    }
    
    public synchronized ByteBuffer append(char str[]) {
        return append(String.valueOf(str));
    }
    
    public synchronized ByteBuffer append(byte str[]) {
        int len = str.length;
        int newcount = count + len;
        if (newcount > value.length)
            expandCapacity(newcount);
        System.arraycopy(str, 0, value, count, len);
        count = newcount;
        return this;
    }
    
    public synchronized ByteBuffer append(byte str[], int offset, int len) {
        int newcount = count + len;
        if (newcount > value.length)
            expandCapacity(newcount);
        System.arraycopy(str, offset, value, count, len);
        count = newcount;
        return this;
    }
    
    public ByteBuffer append(boolean b) {
        return append(String.valueOf(b));
    }
    
    public synchronized ByteBuffer append(char c) {
        return append(String.valueOf(c));
    }
    
    public ByteBuffer append(int i) {
        return append(String.valueOf(i));
    }
    
    public ByteBuffer append(long l) {
        return append(String.valueOf(l));
    }
    
    public ByteBuffer append(float f) {
        return append(String.valueOf(f));
    }
    
    public ByteBuffer append(double d) {
        return append(String.valueOf(d));
    }
    
    public synchronized ByteBuffer delete(int start, int end) {
        if (start < 0)
            throw new StringIndexOutOfBoundsException(start);
        if (end > count)
            end = count;
        if (start > end)
            throw new StringIndexOutOfBoundsException();
        
        int len = end - start;
        if (len > 0) {
            if (shared)
                copy();
            System.arraycopy(value, start+len, value, start, count-end);
            count -= len;
        }
        return this;
    }
    
    public synchronized ByteBuffer deleteCharAt(int index) {
        if ((index < 0) || (index >= count))
            throw new StringIndexOutOfBoundsException();
        if (shared)
            copy();
        System.arraycopy(value, index+1, value, index, count-index-1);
        count--;
        return this;
    }
    
    public synchronized ByteBuffer replace(int start, int end, String str) {
        if (start < 0)
            throw new StringIndexOutOfBoundsException(start);
        if (end > count)
            end = count;
        if (start > end)
            throw new StringIndexOutOfBoundsException();
        
        byte str_byte[] = str.getBytes();
        
        int len = str_byte.length;
        int newCount = count + len - (end - start);
        if (newCount > value.length)
            expandCapacity(newCount);
        else if (shared)
            copy();
        
        System.arraycopy(value, end, value, start + len, count - end);
        
        for(int i = 0; i < str_byte.length; i++)
            value[start+i] = str_byte[i];
        
        count = newCount;
        
        return this;
    }
    
    public String substring(int start) {
        return substring(start, count);
    }
    
    public synchronized String substring(int start, int end) {
        if (start < 0)
            throw new StringIndexOutOfBoundsException(start);
        if (end > count)
            throw new StringIndexOutOfBoundsException(end);
        if (start > end)
            throw new StringIndexOutOfBoundsException(end - start);
        return new String(value, start, end - start);
    }
    
    public synchronized ByteBuffer insert(int index, byte str[], int offset,
                                          int len) {
        if ((index < 0) || (index > count))
            throw new StringIndexOutOfBoundsException();
        if ((offset < 0) || (offset + len > str.length))
            throw new StringIndexOutOfBoundsException(offset);
        if (len < 0)
            throw new StringIndexOutOfBoundsException(len);
        int newCount = count + len;
        if (newCount > value.length)
            expandCapacity(newCount);
        else if (shared)
            copy();
        System.arraycopy(value, index, value, index + len, count - index);
        System.arraycopy(str, offset, value, index, len);
        count = newCount;
        return this;
    }
    
    public synchronized ByteBuffer insert(int offset, Object obj) {
        return insert(offset, String.valueOf(obj));
    }
    
    public synchronized ByteBuffer insert(int offset, String str) {
        if ((offset < 0) || (offset > count)) {
            throw new StringIndexOutOfBoundsException();
        }
        
        if (str == null) {
            str = String.valueOf(str);
        }
        
        byte str_byte[] = str.getBytes();
        
        return insert(offset, str_byte, 0, str_byte.length);
    }
    
    public synchronized ByteBuffer insert(int offset, byte str[]) {
        return insert(offset, str, 0, str.length);
    }
    
    public synchronized ByteBuffer insert(int offset, char str[]) {
        return insert(offset, String.valueOf(str));
    }
    
    public ByteBuffer insert(int offset, boolean b) {
        return insert(offset, String.valueOf(b));
    }
    
    public synchronized ByteBuffer insert(int offset, char c) {
        return insert(offset, String.valueOf(c));
    }
    
    public ByteBuffer insert(int offset, int i) {
        return insert(offset, String.valueOf(i));
    }
    
    public ByteBuffer insert(int offset, long l) {
        return insert(offset, String.valueOf(l));
    }
    
    public ByteBuffer insert(int offset, float f) {
        return insert(offset, String.valueOf(f));
    }
    
    public ByteBuffer insert(int offset, double d) {
        return insert(offset, String.valueOf(d));
    }
    
    public synchronized ByteBuffer reverse() {
        if (shared) copy();
        int n = count - 1;
        for (int j = (n-1) >> 1; j >= 0; --j) {
            byte temp = value[j];
            value[j] = value[n - j];
            value[n - j] = temp;
        }
        return this;
    }
    
    public String toString() {
        return new String(value);
    }
    
    final void setShared() { shared = true; } 
    final byte[] getValue() { return value; }
    
    private void readObject(java.io.ObjectInputStream s)
        throws java.io.IOException, ClassNotFoundException {
        s.defaultReadObject();
        value = (byte[]) value.clone();
        shared = false;
    }
}
