/* 
 * Program:             $RCSfile: SpdEncoding.java,v $
 * 
 * Purpose:             
 * 
 * Author:              T.Shikimachi  2003/10/03
 *                      S.Yamamoto  2003/10/03
 * 
 * (C) Copyright:       T.Shikimachi and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: SpdEncoding.java,v 50.5 2003/10/12 05:31:49 yamamoto Exp yamamoto $
 */

package sapid.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import sun.io.MalformedInputException;

import sapid.util.SpdMsg;

/**
 *  Wrapper class of encoding
 *
 *  @author T.Shikimachi, H.Kuwabara
 */

public final class SpdEncoding {
    /* Preferred MIME Name */
    private static final String CODE_EUCJP = "EUC-JP";
    private static final String CODE_SJIS = "Shift_JIS";
    private static final String CODE_JIS = "ISO-2022-JP";
    private static final String CODE_ASCII = "US-ASCII";
    private static final String CODE_UTF8 = "UTF-8";
    // private static final String CODE_UTF16 = "UTF-16";
    // private static final String CODE_UNICODE_B = "UTF-16B";
    // private static final String CODE_UNICODE_L = "UTF-16L";
    // private static final String CODE_UNICODE_BU = "UTF-16BU";
    // private static final String CODE_UNICODE_LU = "UTF-16LU";
    private static final String CODE_UNKNOWN = "UNKNOWN";
    private static final String CODE_NONE = "NOT SPECIFIED";

    /* Sapid Encoding class (constant) */
    public static final SpdEncoding EUCJP = new SpdEncoding(CODE_EUCJP);
    public static final SpdEncoding SJIS = new SpdEncoding(CODE_SJIS);
    public static final SpdEncoding JIS = new SpdEncoding(CODE_JIS);
    public static final SpdEncoding ASCII = new SpdEncoding(CODE_ASCII);
    public static final SpdEncoding UTF8 = new SpdEncoding(CODE_UTF8);
    // public static final SpdEncoding UTF16 = new SpdEncoding(CODE_UTF16);
    // public static final SpdEncoding UNICODE_B = new SpdEncoding(CODE_UNICODE_B);
    // public static final SpdEncoding UNICODE_L = new SpdEncoding(CODE_UNICODE_L);
    // public static final SpdEncoding UNICODE_BU = new SpdEncoding(CODE_UNICODE_BU);
    // public static final SpdEncoding UNICODE_LU = new SpdEncoding(CODE_UNICODE_LU);
    public static final SpdEncoding UNKNOWN = new SpdEncoding(CODE_UNKNOWN);
    public static final SpdEncoding NONE = new SpdEncoding(CODE_NONE);

    /** upper bound for checked string length */
    private static final int UPPER_BOUND = 4096;

    private String encodingName;

    private SpdEncoding() {
    }

    /**
     *  @param encodingName 
     */
    private SpdEncoding(String encodingName) {
	this.encodingName = encodingName;
    }

    /**
     *  Get file encoding
     *  @param file 
     *  @return 
     */
    public static SpdEncoding getEncoding(File file) 
	throws FileNotFoundException {
	if (JIS.compare(file)) {
	    return JIS;
	}
	if (EUCJP.compare(file)) {
	    return EUCJP;
	}
	if (SJIS.compare(file)) {
	    return SJIS;
	}
	if (ASCII.compare(file)) {
	    return ASCII;
	}

	return UNKNOWN;
    }

    /**
     *  Get file encoding
     *  @param path
     *  @return 
     */
    public static SpdEncoding getEncoding(String path) 
	throws FileNotFoundException {
	return getEncoding(new File(path));
    }

    /**
     *  Compare encoding
     *  @param file 
     *  @return 
     */
    private boolean compare(File file) 
	throws FileNotFoundException {
	FileInputStream fisSource = new FileInputStream(file);
	FileInputStream fisTarget = new FileInputStream(file);

	try {
	    InputStreamReader isrSource = new InputStreamReader(fisSource, 
								encodingName);
	    BufferedReader brSource = new BufferedReader(isrSource);

	    InputStreamReader isrTarget = new InputStreamReader(fisTarget, 
								"JISAutoDetect");
	    BufferedReader brTarget = new BufferedReader(isrTarget);

	    StringBuffer strSource = new StringBuffer();
	    StringBuffer strTarget = new StringBuffer();
	    String strSourceLine = brSource.readLine();
	    String strTargetLine = brTarget.readLine();

	    while (strSourceLine != null && strTargetLine != null) {
		boolean isMultiByte = hasMultiByteChar(strTargetLine);

		if (isMultiByte && isASCII()) {
		    return false;
		}

		if ((isMultiByte && !encodingName.equals(CODE_ASCII)) ||
		    (!isMultiByte && encodingName.equals(CODE_ASCII))) {
		    strSource.append(strSourceLine);
		    strTarget.append(strTargetLine);
		}

		if (strTarget.length() > UPPER_BOUND) {
		    break;
		}

		strSourceLine = brSource.readLine();
		strTargetLine = brTarget.readLine();
	    } 

	    if (strTarget.length() > 0 && 
		strTarget.toString().equals(strSource.toString())) {
		return true;
	    }
	} catch (sun.io.MalformedInputException e) {
	    SpdMsg.verbose(SpdEncoding.class.getName() + ": " + e.toString());
	} catch (UnsupportedEncodingException e) {
	    e.printStackTrace();
	} catch (FileNotFoundException e) {
	    throw e;
	} catch (IOException e) {
	    e.printStackTrace();
	} finally {
	    try {
		fisSource.close();
	    } catch (IOException e) {
	    }

	    try {
		fisTarget.close();
	    } catch (IOException e) {
	    }
	}

	return false;
    }

    /**
     *  Get encoding attribute's value in XML declaration.
     *  @param file
     *  @return 
     */
    public static SpdEncoding getEncodingInXMLDecl(File file)
	throws FileNotFoundException {
	FileReader fr = new FileReader(file);

	try {
	    BufferedReader br = new BufferedReader(fr);

	    String strLine = br.readLine();
	    if (strLine.startsWith("<?xml")) {
		int index = strLine.indexOf("encoding");
		if (index >= 0) {
		    int start = strLine.indexOf("\"", index) + 1;
		    int end = strLine.indexOf("\"", start);
		    return newInstance(strLine.substring(start, end));
		}
	    }
	} catch (FileNotFoundException e) {
	    throw e;
	} catch (IOException e) {
	    e.printStackTrace();
	} finally {
	    try {
		fr.close();
	    } catch (IOException e) {
	    }
	}

	return UNKNOWN;
    }

    /**
     *  Get encoding attribute's value in XML declaration.
     *  @param path
     *  @return 
     */
    public static SpdEncoding getEncodingInXMLDecl(String path)
	throws FileNotFoundException {
	return getEncodingInXMLDecl(new File(path));
    }

    /**
     *  Check equality
     *  @param encoding 
     *  @return 
     */
    public boolean equals(SpdEncoding encoding) {
	// return (this == encoding);
	return encodingName.equals(encoding.toString());
    }

    public boolean isEUC() {
	return equals(EUCJP);
    }

    public boolean isShiftJIS() {
	return equals(SJIS);
    }

    public boolean isJIS() {
	return equals(JIS);
    }

    public boolean isASCII() {
	return equals(ASCII);
    }

    /**
     *  SpdEncoding to String (to get Preferred MIME Name)
     *  @return 
     */
    public String toString() {
	return encodingName;
    }

    /**
     *  Create new object that has specified encoding name.
     *  @param encodingName
     *  @return 
     */
    public static SpdEncoding newInstance(String encodingName) {
	return (new SpdEncoding(encodingName));
    }

    /**
     *  Check that given string has multibyte char or not.
     *  @param string
     *  @return
     */
    private static boolean hasMultiByteChar(String string) {
	for (int i = 0; i < string.length(); i++) {
	    if (String.valueOf(string.charAt(i)).getBytes().length > 1) {
		return true;
	    }
	}
	return false;
    }

    public static void main(String[] args) 
	throws FileNotFoundException {
	SpdMsg.setVerboseMode(true);
	System.out.println(getEncoding(new File(args[0])));
	// System.out.println(getEncodingInXMLDecl(new File(args[0])));
    }
}
