/* 
 * Program:             $RCSfile: SpdLineSeparator.java,v $
 * 
 * Purpose:             Wrapper class of line separator.
 * 
 * Author:              H.Kuwabara  2003/10/14
 *                      S.Yamamoto  2003/09/04
 * 
 * (C) Copyright:       H.Kuwabara and S.Yamamoto  2003
 *                      This file is a product of the project Sapid.
 */

/* 
 * $Id: SpdLineSeparator.java,v 50.2 2003/10/14 15:00:04 yamamoto Exp yamamoto $
 */

package sapid.util;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;

/**
 *  Wrapper class of line separator.
 *
 *  @author H.Kuwabara
 */

public final class SpdLineSeparator {
    private static final String CODE_CRLF = "\r\n";
    private static final String CODE_LF = "\n";
    private static final String CODE_CR = "\r";
    private static final String CODE_NONE = "NONE";

    public static final SpdLineSeparator CRLF = new SpdLineSeparator(CODE_CRLF);
    public static final SpdLineSeparator LF = new SpdLineSeparator(CODE_LF);
    public static final SpdLineSeparator CR = new SpdLineSeparator(CODE_CR);
    public static final SpdLineSeparator NONE = new SpdLineSeparator(CODE_NONE);

    private String codeLineSeparator;

    private SpdLineSeparator() {
    }

    /**
     *  Constructs a new SpdLineSeparator with a given line separator.
     *  @param codeLineSeparator a line separator
     */
    private SpdLineSeparator(String codeLineSeparator) {
        this.codeLineSeparator = codeLineSeparator;
    }

    /**
     *  Get line separator which is used in given file.
     *  @param file
     */
    public static SpdLineSeparator getLineSeparator(File file) 
        throws FileNotFoundException {
        FileInputStream fis = new FileInputStream(file);
        BufferedInputStream bis = new BufferedInputStream(fis);

        try {
            int byteLF = (int)CODE_LF.getBytes()[0];
            int byteCR = (int)CODE_CR.getBytes()[0];

            int bytecode = bis.read();
            while(bytecode != -1) {
                if (bytecode == byteLF) {
                    return LF;
                } else if (bytecode == byteCR) {
                    bytecode = bis.read();
                    if (bytecode == byteLF) {
                        return CRLF;
                    } else {
                        return CR;
                    }
                }
                bytecode = bis.read();
            }
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                fis.close();
            } catch (IOException e) {
            }
        }

        return NONE;
    }

    /**
     *  Get line separator which is used in given file.
     *  @param path
     */
    public static SpdLineSeparator getLineSeparator(String path) 
        throws FileNotFoundException {
        return getLineSeparator(new File(path));
    }

    /**
     *  Check equality
     *  @param codeLineSeparator
     *  @return 
     */
    public boolean equals(SpdLineSeparator lineSeparator) {
        //return (this == lineSeparator);
        return codeLineSeparator.equals(lineSeparator.toString());
    }

    public boolean isCRLF() {
        return equals(CRLF);
    }

    public boolean isLF() {
        return equals(LF);
    }

    public boolean isCR() {
        return equals(CR);
    }

    public boolean isNONE() {
        return equals(NONE);
    }

    /**
     *  SpdLineSeparator to String
     *  @return
     */
    public String toString() {
        return codeLineSeparator;
    }

    public static void main(String[] args) 
        throws FileNotFoundException {
        SpdLineSeparator lineSeparator = SpdLineSeparator.getLineSeparator(args[0]);
        System.err.println("CR+LF : " + lineSeparator.isCRLF());
        System.err.println("LF    : " + lineSeparator.isLF());
        System.err.println("CR    : " + lineSeparator.isCR());
        System.err.println("NONE  : " + lineSeparator.isNONE());
    }
}
