/* 
 * Program:             $RCSfile: spdDetectEncoding.c,v $  $Revision: 40.8 $
 * 
 * Purpose:
 *
 * Original author:     Ken R. Lunde  1993
 *
 * Modified:            S.Yamamoto  2003/01/08
 *
 * (C) Copyright:       Ken R. Lunde and S.Yammoto  1993 - 2003
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdDetectEncoding.c,v 40.8 2003/03/27 15:23:50 yamamoto Exp yamamoto $";
#endif

/*
Program: jcode.c
Version: 3.0
Date:    July 1, 1993
Author:  Ken R. Lunde, Adobe Systems Incorporated
  EMAIL: lunde@mv.us.adobe.com
  MAIL : 1585 Charleston Road, P.O. Box 7900, Mountain View, CA 94039-7900
Type:    A tool for displaying the electronic values of Japanese characters.
Code:    ANSI C (portable)

PORTABILITY:
This source code was written so that it would be portable on C compilers which
conform to the ANSI C standard. It has been tested on a variety of compilers.

I used THINK C and GNU C as my development platforms. I left in the Macintosh-
specific lines of code so that it would be easier to enhance/debug this tool
later. For those of you who wish to use this tool on the Macintosh, simply
add the ANSI library to the THINK C project, and then build the application.
Be sure that THINK_C has been defined, though, as the conditional compilation
depends on it. You then have a double-clickable application, which when
launched, will greet you with a Macintosh-style interface.

DISTRIBUTION AND RESTRICTIONS ON USAGE:
 1) Please give this source code away to your friends at no charge.
 2) Please try to compile this source code on various platforms to check for
    portablity, and please report back to me with any results be they good or
    bad. Suggestions are always welcome.
 3) Only use this tool on a copy of a file -- do not use an original. This
    is just common sense.
 4) This source code or a compiled version may be bundled with commercial
    software as long as the author is notified beforehand. The author's name
    should also be mentioned in the credits.
 5) Feel free to use any of the algorithms for your own work. Many of them are
    being used in other tools I have written.
 6) The most current version can be obtained by requesting a copy directly
    from me.

DESCRIPTION:
 1) Supports Shift-JIS, SPD_EC_EUC, New-JIS, Old-JIS, and SPD_EC_NEC_JIS-JIS codes.
 2) Automatically detects infile's Japanese code.
 3) The tool reads the infile, and will create an outfile which
    contains all of the characters within the infile along with their
    electronic codes. The user may choose octal, decimal, or hexadecimal
    notation (the default, if none specified, is hexadecimal).
 4) Supports half-width katakana in both Shift-JIS and SPD_EC_EUC codes as input.
 5) The outfile specifies the Shift-JIS, SPD_EC_EUC, JIS, SPD_EC_ASCII, and KUTEN values
    for each character in the file. The default file extension is ".val" if
    no outfile is specified.

FILE HANDLING:
 1) Specifying the infile is optional as one can redirect stdin and stdout.
 2) Specifying the outfile is also optional. If none is specified, the
    tool will semi-intelligently change the file's name. The tool
    simply scans the outfile name, finds the last period in it, terminates
    the string at that point, and tacks on an extension. Here are some example
    command lines, and the resulting outfile names:

    a) jcode sig.jpn                      = sig.val
    b) jcode sig.jpn sig.out              = sig.out
    c) jcode sig.jpn.txt                  = sig.jpn.val
    d) jcode sig                          = sig.val

    This is very useful for MS-DOS users since a filename such as sig.jpn.val
    will not result in converting a file called sig.jpn.

    Also note that if the outfile and infile have the same name, the tool
    will not work, and data will be lost. I tried to build safe-guards against
    this. For example, note how this tool will change the outfile name so
    that it does not overwrite the infile:
    
    a) jcode sig.val                      = sig-.val
    b) jcode sig.val sig.val              = sig-.val
    c) jcode sig-.val                     = sig--.val
    
    If only the infile is given, a hyphen is inserted after the last period,
    and the extension is then reattached. If the outfile is specified by the
    user, then it will search for the last period (if any), attach a hyphen,
    and finally attach the proper extension). This sort of protection is NOT
    available from this tool if stdin/stdout are used.
 3) If you want to specify an infile, but to have the output to go to stdout,
    you have two options:

    a) jcode < sig.jpn                    <= redirect stdin for infile
    b) jcode sig.jpn -                    <= specify infile, then use "-" to
                                             indictate stdout for output

 4) When using pipes under UNIX, it is very important to use the "-iCODE"
    option so that the automatic code detection routines are skipped. This
    is because the tool attempts to use the rewind() function, and this
    has no effect on some machines if input is coming from a pipe. Here
    are some examples (they assume that the infile code is SPD_EC_EUC):

    a) cat japan.txt | jcode -oj          <= does not work
    a) cat japan.txt | jcode -ie -oj      <= works
*/

#include <ctype.h>
#include <stdio.h>
#include <Sapid/Sapid.h>

/*
 * Note by S.Yamamoto 2003/01/08.
 * 
 * (1) These macros are deleted because they aren't used in function
 *     spdDetectEncodingJa().
 * 
 * #define NOTSET       0
 * #define KUTEN        8
 * #define JISORSJIS    9
 *
 * (2) Function name DetectCodeType() is renamed as spdDetectEncodingJa().
 * 
 * (3) Constant names for encoding type are changed to have "SPD_EC_" prefix
 *     in accrodance with Sapid convention.
 */

static SpdEncode spdDetectEncodingJa(FILE *in)
{
  int c = 0,whatcode = SPD_EC_ASCII;

  while ((whatcode == SPD_EC_EUCORSJIS || whatcode == SPD_EC_ASCII) && c != EOF) {
    if ((c = fgetc(in)) != EOF) {
      if (c == '\033') {
        c = fgetc(in);
        if (c == '$') {
          c = fgetc(in);
          if (c == 'B')
            whatcode = SPD_EC_JIS;
          else if (c == '@')
            whatcode = SPD_EC_JIS;
        }
        else if (c == 'K')
          whatcode = SPD_EC_JIS;
      }
      else if ((c >= 129 && c <= 141) || (c >= 143 && c <= 159))
        whatcode = SPD_EC_SJIS;
      else if (c == 142) {
        c = fgetc(in);
        if ((c >= 64 && c <= 126) || (c >= 128 && c <= 160) || (c >= 224 && c <= 252))
          whatcode = SPD_EC_SJIS;
        else if (c >= 161 && c <= 223)
          whatcode = SPD_EC_EUCORSJIS;
      }
      else if (c >= 161 && c <= 223) {
        c = fgetc(in);
        if (c >= 240 && c <= 254)
          whatcode = SPD_EC_EUC;
        else if (c >= 161 && c <= 223)
          whatcode = SPD_EC_EUCORSJIS;
        else if (c >= 224 && c <= 239) {
          whatcode = SPD_EC_EUCORSJIS;
          while (c >= 64 && c != EOF && whatcode == SPD_EC_EUCORSJIS) {
            if (c >= 129) {
              if (c <= 141 || (c >= 143 && c <= 159))
                whatcode = SPD_EC_SJIS;
              else if (c >= 253 && c <= 254)
                whatcode = SPD_EC_EUC;
            }
            c = fgetc(in);
          }
        }
        else if (c <= 159)
          whatcode = SPD_EC_SJIS;
      }
      else if (c >= 240 && c <= 254)
        whatcode = SPD_EC_EUC;
      else if (c >= 224 && c <= 239) {
        c = fgetc(in);
        if ((c >= 64 && c <= 126) || (c >= 128 && c <= 160))
          whatcode = SPD_EC_SJIS;
        else if (c >= 253 && c <= 254)
          whatcode = SPD_EC_EUC;
        else if (c >= 161 && c <= 252)
          whatcode = SPD_EC_EUCORSJIS;
      }
    }
  }
  return whatcode;
}

/******************************************************************************/

SpdEncode spdDetectEncoding(FILE *in)
{
    long        saved_pos;
    SpdEncode   encoding;

    saved_pos = ftell(in);
    /* spdMessage("spdDetectEncoding(): ftell() %d.\n", ftell(in)); */
    fseek(in, 0, SEEK_SET);
    /* spdMessage("spdDetectEncoding(): ftell() %d.\n", ftell(in)); */
    encoding = spdDetectEncodingJa(in);
    fseek(in, saved_pos, SEEK_SET);
    return (encoding);
}
