/* 
 * Program:             $RCSfile: spdEncode.c,v $  $Revision: 40.30 $
 * 
 * Purpose:             
 * 
 * Author:              A.Yoshida  2002/05/21
 *                      S.Yamamoto  2002/05/21
 * 
 * (C) Copyright:       A.Yoshida and S.Yamamoto  2002 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdEncode.c,v 40.30 2003/07/06 10:09:00 yamamoto Exp yamamoto $";
#endif

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <Sapid/Sapid.h>
#include <Sapid/spdSet2.h>

/* #define SPD_ENCODE_DEBUG */

#define MAX_ENC_LEN         16          /* Maximum length of encode name. */

SpdEncodeAll    spdEncode = {
    SPD_EC_ASCII
};

static SpdEncode spdStrToEncode(SpdString enName);
static void spdLoadEncodeNameTable_str2enc(void);
static void spdLoadEncodeNameTable_enc2str(void);
static SpdString lower_name(SpdString str);
static void spdUnloadEncodeNameTable_all(void);
static void set_unload(void);

typedef struct {
    SpdString   name;
    int         id;
} EncodeName;

/*
 * Note: The name of a encode which SpdEncodeToStr() returns is
 * the first one in the list of the same encode.
 */

static EncodeName encode_name[] = {
    {"ASCII",       SPD_EC_ASCII},
    {"US-ASCII",    SPD_EC_ASCII},          /* Preferred MIME name. */
    {"ISO646-US",   SPD_EC_ASCII},
    {"C",           SPD_EC_ASCII},
    
    {"EUC-JP",      SPD_EC_EUC},            /* Preferred MIME name. */
    {"EUC_JP",      SPD_EC_EUC},            /* Java encoding name. */
    {"EUC",         SPD_EC_EUC},
    {"ja_JP.euc",   SPD_EC_EUC},
    {"ja_JP.eucJP", SPD_EC_EUC},

    {"ISO-8859-1",  SPD_EC_EUCORSJIS},      /* Preferred MIME name. */
    {"ISO8859_1",  SPD_EC_EUCORSJIS},       /* Java encoding name. */
    
    {"ISO-2022-JP", SPD_EC_JIS},            /* Preferred MIME name. */
    {"ISO2022JP", SPD_EC_JIS},              /* Java encoding name. */
    {"JIS",         SPD_EC_JIS},
    /* !! Locale style. */
    
    {"Shift_JIS",   SPD_EC_SJIS},           /* Preferred MIME name. */
    {"SJIS",        SPD_EC_SJIS},           /* Java encoding name. */
    {"ja_JP.sjis",  SPD_EC_SJIS},

    {NULL,          SPD_EC_UNKNOWN}
};

static SpdSet2  str2enc_tbl = NULL;
static SpdSet2  enc2str_tbl = NULL;

SpdBoolean spdInputEncodeIsWChar(unsigned char c1, unsigned char c2)
{
    return spdEncodeIsWChar(spdEncode.in, c1, c2);
}

SpdBoolean spdEncodeIsWChar(SpdEncode en, unsigned char c1, unsigned char c2)
{
    SpdBoolean  res;
    
    switch (en) {
      case SPD_EC_ASCII:
        res = FALSE;                    /* ASCII has no wide character. */
        break;

      case SPD_EC_EUC:
          res = (c1 > 0xA1 && c2 > 0xA1);       /* JIS X 0208 */
          switch (c1) {
          case 0x8E: /* SS2: single shift 2, JIS X 0201 Hankaku Kana */
              res = TRUE;
              break;
          case 0x8F: /* SS3: single shift 3, JIS X 0212 Hozyo Kanji */
              {
                  static SpdBoolean first_SS3 = FALSE;
                  if (!first_SS3) {
                      spdWarning("spdEncodeIsWChar(): "
                                 "JIS X 0212 is not supported.\n");
                      first_SS3 = TRUE;
                  }
              }
              break;
          case 0x1B: /* ESC may be the first code of locking shift. */
              {
                  static SpdBoolean first_LS = FALSE;
                  switch (c2) {
                  case 0x6E: /* LS2 */  case 0x6F: /* LS3 */
                  case 0x7E: /* LS1R */ case 0x7D: /* LS2R */
                  case 0x7C: /* LS3R */
                      if (!first_LS) {
                          spdWarning("spdEncodeIsWChar(): "
                                     "Locking shift of EUC is not supported.\n");
                      }
                      first_LS = TRUE;
                  }
              }
              break;
          } /* end of switch(c1) */
          break;
      case SPD_EC_EUCORSJIS:
          res = spdEncodeIsWChar(SPD_EC_EUC, c1, c2) 
              || spdEncodeIsWChar(SPD_EC_SJIS, c1, c2);
          break;
      case SPD_EC_JIS:
        res = FALSE;                         /* !! Temporarily. */
        /* ISO-2022JP uses locking shift mechanism. */
        /*      ESC ( B  ASCII
         *      ESC ( J  JIS X 0201
         *      ESC $ @  JIS X 0208-1978
         *      ESC $ B  JIS X 0208-1983
         */
        break;
        
      case SPD_EC_SJIS:
        res = ((0201 <= c1 && c1 <= 0237) || (0340 <= c1 && c1 <= 0357)) &&
              ((0100 <= c2 && c2 <= 0176) || (0200 <= c2 && c2 <= 0374));
        break;
      default:
        spdErrorExit("spdEncodeIsWChar(): Unknown Encode: %d for \"%x%x\".\n",
                     en, c1, c2);
        SPD_DUMMY_RETURN(FALSE);
        break;
    }

    return res;
}

SpdBoolean spdInputEncodeIs(SpdEncode en)
{
    return (spdEncode.in == en);
}

void spdSetInputEncode(SpdString enName)
{
    if ((spdEncode.in = spdStrToEncode(enName)) == SPD_EC_UNKNOWN) {
        spdErrorExit("spdStrToEncode(): Unknown encode name \"%s\".\n",
                     enName);
    }
    switch (spdEncode.in) {
    case SPD_EC_JIS:
        spdWarning("spdSetInputEncode(): %s is not supported.\n", enName);
        break;
    default:
        break;
    }
}

static SpdEncode spdStrToEncode(SpdString enName)
{
    SpdEncode   res = SPD_EC_UNKNOWN;
    SpdElem     *el;

    spdLoadEncodeNameTable_str2enc();
    if ((el = spdSearchElemS(str2enc_tbl, lower_name(enName))) != NULL) {
        res = spdSetGetValueI(el->value);
    }
    return res;
}

SpdString spdEncodeToStr(SpdEncode encode)
{
    static char res[MAX_ENC_LEN];
    SpdElem     *el;
    
    spdLoadEncodeNameTable_enc2str();
    if ((el = spdSearchElemI(enc2str_tbl, encode)) != NULL) {
        strncpy(res, spdSetGetValueS(el->value), MAX_ENC_LEN);
    } else {
        res[0] = '\0';
    }
    return res;
}

static SpdString lower_name(SpdString str)
{
    static char name[MAX_ENC_LEN + 1];
    int         i;
    
    for (i = 0; i <= MAX_ENC_LEN && str[i] != '\0'; i++) {
        name[i] = tolower(str[i]);
    }
    name[i] = '\0';
    return name;
}

static void spdUnloadEncodeNameTable_all(void)
{
    spdFreeSet2(str2enc_tbl);
    spdFreeSet2(enc2str_tbl);
}

static void set_unload(void)
{
    static int unload = FALSE;

    if (!unload && atexit(spdUnloadEncodeNameTable_all) < 0) {
        spdErrorExit("spdLoadEncodeNameTable(): atexit() fails: %s", 
                     strerror(errno));
    } else {
        unload = TRUE;
    }
}

static void spdLoadEncodeNameTable_str2enc(void)
{
    int     i;

    set_unload();

    if (str2enc_tbl == NULL) {
        str2enc_tbl = spdCreateSet2(SPDSET_COPY_STRING_KEY, SPDSET_INT_VALUE);
    }

    for (i = 0; encode_name[i].name != NULL; i++) {
        if (spdAddElemSI(str2enc_tbl, lower_name(encode_name[i].name),
                         encode_name[i].id) == NULL) {
            spdErrorExit("spdLoadEncodeNameTable_str2enc(): Can't add an element to str2enc_tbl.");
        }
    }

#ifdef SPD_ENCODE_DEBUG
    fprintf(stderr, "DEBUG: spdLoadEncodeNameTable_str2enc()\n");
    for (i = 0; encode_name[i].name != NULL; i++) {
        SpdElem     *el;
        
        if ((el = spdSearchElemS(str2enc_tbl, encode_name[i].name)) != NULL) {
            fprintf(stderr, "DEBUG:   %s -> %d\n", encode_name[i].name, 
                    spdSetGetValueI(el->value));
        }
    }
#endif
}

static void spdLoadEncodeNameTable_enc2str(void)
{
    int     i;

    set_unload();

    if (enc2str_tbl == NULL) {
        enc2str_tbl = spdCreateSet2(SPDSET_INT_KEY, SPDSET_STRING_VALUE);
    }
    for (i = 0; encode_name[i].name != NULL; i++) {
        if (spdContainsI(enc2str_tbl, encode_name[i].id)) {
            /* The id has been registered and is to be ignored. */
            continue;
        }
        if (spdAddElemIS(enc2str_tbl, 
                         encode_name[i].id, encode_name[i].name) == NULL) {
            spdErrorExit("spdLoadEncodeNameTable_enc2str(): Can't add element to str2enc_tbl.");
        }
    }

#ifdef SPD_ENCODE_DEBUG
    fprintf(stderr, "DEBUG: spdLoadEncodeNameTable_enc2str()\n");
    for (i = 0; encode_name[i].name != NULL; i++) {
        SpdElem     *el;
        
        if ((el = spdSearchElemI(enc2str_tbl, encode_name[i].id)) != NULL) {
            fprintf(stderr, "DEBUG:   %d -> %s\n", encode_name[i].id,
                    spdSetGetValueS(el->value));
        }
    }
#endif
}
