/* 
 * Program:             $RCSfile: spdOpenSDBfile.c,v $  $Revision: 30.51 $
 * 
 * Purpose:
 * 
 * Author:              R.Miura  1998/12/05
 *                      S.Yamamoto  1998/12/05
 * 
 * (C) Copyright:       R.Miura and S.Yamamoto  1998 - 2003
 *                      This file is a product of the project Sapid.
 */

#ifndef lint
static char rcsid[] =
    "$Id: spdOpenSDBfile.c,v 30.51 2003/08/16 10:32:28 yamamoto Exp yamamoto $";
#endif

#include <stdio.h>
#include <Sapid/Sapid.h>

static FILE *spdOpenFile(const char *fileName, int mode);

/* 
 * Note:
 * 
 * (1) Value of the third argument "int mode" for spdOpenSDBfile() is
 * one of SAPID_O_{READ,WRITE,APPEND}.
 * 
 * (2) Value of the second argument "char *type" for spdFopen() is
 * one of SPD_FOPEN_[RWA][TB]_MODE.
 */

/* 
 * Check path and make directory for write.
 */

static SpdBoolean checkSDBpath(const char *dirname, int mode)
{
    if (!spdIsDirectory(dirname)) {
        if (mode == SAPID_O_READ || mode == SAPID_O_APPEND) {
            return (FALSE);
        } else if (mode == SAPID_O_WRITE) {
            return (spdMkdirR(dirname));
        }
    }
    return (TRUE);
}

/* 
 * Open SDB file.
 */

static FILE *spdOpenFile(const char *fileName, int mode)
{
    int     format;

    /* 
     * type[0] = [rwa] (read, write, append).
     * type[1] = [bt] (binary, text).
     * type[2] = '\0' (terminator).
     */
    char type[3];

#ifdef SPD_DUMP_READ
    spdMessage("spdOpenFile(): fileName \"%s\", mode %d.\n", fileName, mode);
#endif  

    if (fileName == NULL) {
        return (NULL);
    }

    /* 
     * Set open mode.
     */
    switch (mode) {
      case SAPID_O_READ:
        type[0] = 'r';
        format = spdGetReadSDBformat();
        break;
      case SAPID_O_WRITE:
        type[0] = 'w';
        format = spdGetWriteSDBformat();
        break;
      case SAPID_O_APPEND:
        type[0] = 'a';
        format = spdGetWriteSDBformat();
        break;
      default:
        spdErrorExit("spdOpenFile(): %d: Invalid open mode.\n", mode);
        SPD_DUMMY_RETURN(NULL);
        break;
    }

    /* 
     * Set format mode.
     */
    switch (format) {
      case SAPID_BINARY_FORMAT_SDB:
        type[1] = 'b';
        break;
      case SAPID_TEXT_FORMAT_SDB:

#ifdef WIN32
        type[1] = 't';                      /* Windows. */
#else
        type[1] = '\0';                     /* Almost all UNIXes. */
#endif

        break;
      default:
        spdErrorExit("spdOpenFile(): %d: Invalid format.\n", format);
    }

    /* 
     * Set terminator.
     */
    type[2] = '\0';

    return (spdFopen(fileName, type));
}

static const char *spdMakeSDBpath(const char *prefix, const char *fname,
                                  int mode)
{
    static char     fileName[FILENAME_MAX + 1];
    
    strcpy(fileName, spdGetSDBdir());
    if (fileName[strlen(fileName) - 1] != '/') {
        strcat(fileName, "/");
    }
    strcat(fileName, prefix);
    if (checkSDBpath(fileName, mode) == FALSE) {
        return (NULL);
    }
    if (fileName[strlen(fileName) - 1] != '/') {
        strcat(fileName, "/");
    }
    strcat(fileName, fname);

    return (fileName);
}

/* 
 * Check SDB file.
 */

SpdBoolean spdCheckSDBClassFile(const char *prefix, const char *className)
{
    return (spdIsFile(spdMakeSDBpath(prefix, className, SAPID_O_READ)));
}

/* 
 * Open file in SDB.
 */

FILE *spdOpenSDBfile(const char *prefix, const char *fname, int mode)
{
    return (spdOpenFile(spdMakeSDBpath(prefix, fname, mode), mode));
}

/* 
 * Open ISDB.
 */

FILE *spdOpenISDB(SpdObjId ifileId, int mode)
{
    return (spdOpenSDBfile(ISDB_DIR, spdObjIdStr(ifileId), mode));
}

/* 
 * Open file in SRC.
 */

FILE *spdOpenSourceById(SpdObjId fileId, int mode)
{
    char type[3];

    /* 
     * Set open mode.
     */
    switch (mode) {
      case SAPID_O_READ:
        type[0] = 'r';
        break;
      case SAPID_O_WRITE:
        type[0] = 'w';
        break;
      case SAPID_O_APPEND:
        type[0] = 'a';
        break;
      default:
        spdErrorExit("spdOpenSourceById(): %d: Invalid open mode.\n", mode);
        SPD_DUMMY_RETURN(NULL);
        break;
    }
    type[1] = 'b';
    type[2] = '\0';

    return (spdFopen(spdMakeSDBpath(SAPID_SRC_DIR, spdObjIdStr(fileId), mode), 
                     type));
}

/* 
 * Open Java source.
 */

FILE *spdOpenJavaSource(SpdConstString fileName, int mode)
{
    return (spdOpenFile(fileName, mode));
}

/* 
 * Open Freja.
 */

FILE *spdOpenFreja(const char *prefix, const char *className, int mode)
{
    static char     fileName[FILENAME_MAX + 1];
    
    sprintf(fileName,  "%s/%s", prefix, className);
    return (spdOpenSDBfile(J_MODEL_TMP_DIR, fileName, mode));
}

/******************************************************************************/

#ifdef SPD_MY_FSEEK

SpdOpendFile    spdOpendFile[FOPEN_MAX];    

#endif /* SPD_MY_FSEEK */

/*
 * fopen() function with error checking. Added by S.Yamamoto 2000/02/10.
 */

FILE *spdFopen(const char *fileName, char *type)
{
    FILE        *fp;

    if ((fp = fopen(fileName, type)) == NULL) {
        spdErrorExit("spdFopen(): \"%s\": Can't open file.\n", fileName);
    }

#ifdef SPD_MY_FSEEK
    if (!spdOpendFile[fileno(fp)].isOpened) {
        spdOpendFile[fileno(fp)].isOpened = TRUE;
        spdOpendFile[fileno(fp)].fileId = SPD_NON_ID;
        DW_MY_FSEEK("SPD_MY_FSEEK: Open: %d: spdFopen().\n", fileno(fp));
    }
#endif /* SPD_MY_FSEEK */

    /* spdMessage("Yyy: spdFopen(): \"%s\".\n", fileName); */
    /* spdMessage("Yyy: spdFopen(): %s: \"%s\".\n", type, fileName); */

    return (fp);
}

/*
 * fopen() function without error checking. Added by S.Yamamoto 2000/06/17.
 * Just for tracing open, close and fseek for stream.
 */

FILE *spdSimpleFopen(const char *fileName, char *type)
{
    FILE        *fp;

    if ((fp = fopen(fileName, type)) == NULL) {
        /* spdError("spdFopen(): \"%s\": Can't open file.\n", fileName); */
        /* Nothing. */
    }

#ifdef SPD_MY_FSEEK
    if (fp != NULL && !spdOpendFile[fileno(fp)].isOpened) {
        spdOpendFile[fileno(fp)].isOpened = TRUE;
        spdOpendFile[fileno(fp)].fileId = SPD_NON_ID;
        DW_MY_FSEEK("SPD_MY_FSEEK: Open: %d: spdSimpleFopen().\n", fileno(fp));
    }
#endif /* SPD_MY_FSEEK */

    return (fp);
}

int spdFclose(FILE *fp)
{
    int         retVal;

    if (fp == NULL) {
        return (0);
    }
    
#ifdef SPD_MY_FSEEK
    spdOpendFile[fileno(fp)].isOpened = FALSE;
    DW_MY_FSEEK("SPD_MY_FSEEK:     Close: %d: spdFclose().\n", fileno(fp));
#endif /* SPD_MY_FSEEK */

    if ((retVal = fclose(fp)) != 0) {
        spdError("spdFclose(): fileno %d.\n", fileno(fp));
    }
    return (retVal);
}

/******************************************************************************/

#ifdef DAPID

/*
 * Open DSDB.
 */

FILE *spdOpenDSDB(char *dname, int mode)
{
    return (spdOpenSDBfile(DSDB_DIR, dname, mode));
}

/* 
 * !! ʲ2ĤδؿϤΥե("spdOpenSDBfile.c")٤Ǥ
 * ʤ⤷ʤȤꤢƤ
 * ɥȤڤФҤΥե̾˴ؤؿ
 */

/* 
 * ɥȤΥե̾id饽Ҥifile̾롥 
 */

char *makeDocIfilename(const char *docfilename, const int id)
{
    static char     ifilename[FILENAME_MAX + 1];

    sprintf(ifilename, "%s.%d", docfilename, id);
    return (ifilename);
}

/* 
 * Ǥϡifile̾ˤΥҤޤޤƤsrc_element
 * ֹξƤ롥
 * 
 * :
 * "<docfilename>.<srcelem_objid>"
 */

int getSrcElemObjId(SpdConstString ifilename)
{
    SpdConstString  ep = ifilename + strlen(ifilename);
    SpdConstString  p = ep;
    SpdConstString  sp = NULL;
    char            ostr[FILENAME_MAX + 1];

    /* spdMessage("getSrcElemObjId(): ifilename \"%s\".\n", ifilename); */

    while (--p != ifilename) {
        if (*p == '.') {
            sp = p + 1;
            break;
        }
    }
    if (!sp) {
        spdErrorExit("\"%s\": Can't get srcelem offset.\n", ifilename);
    }

    strncpy(ostr, sp, ep - sp);
    ostr[ep - sp] = '\0';
    /* spdMessage("getSrcElemObjId(): \"%s\": Returns %d.\n",
                  ostr, atoi(ostr)); */
    return (atoi(ostr));
}

#endif /* DAPID */ 

/******************************************************************************/

/*
 * Ԥȥoffset (0Ϥޤ)Ѵ롥
 */

int lineColumnToChar(char *text, int length, int line, int column)
{
    int     l, c, p;
    
    l = 1;
    c = 1;
    p = 0;
    while (TRUE) {
        if (length <= p) {                  /* Error. */
            return (-1);
        } else if (l == line && c == column) {
            break;
        }
        if (text[p] == 0x1b) {              /* Kanji in/out. */
            /* Added by S.Yamamoto 1997/09/14. */
            spdErrorExit("\"%s\": lineColumnToChar(%s, %d, %d, %d)",
                         __FILE__, text, length, line, column);

            if ((text[p+1] == 0x24 && text[p+2] == 0x42) ||
                (text[p+1] == 0x28 && text[p+2] == 0x42)) {
                p += 3;
                continue;
            }
        }
        if (text[p] == '\n') {
            l++;
            c = 0;
        }
        p++;
        c++;
    }
    
    return (p);
}

/*
 * offsetline and columnѴ롥
 */

int charToLineColumn(char *text, int length, int c, int *line, int *column)
{
    int     i;
    
    *line = 1;
    *column = 1;
    if (length <= c) {
        return (-1);
    }

    i = 0;
    while (i < c) {
        if (text[i] == '\n') {
            (*line)++;
            *column = 0;
        }
        i++;
        (*column)++;
    }
    return (0);
}

/*
 * ꤵ줿ԤƬʸoffset롥
 */

char *getHeadOfLine(char *text, int length, int line)
{
    char    *p;
    int     l;
    
    l = 1;
    p = text;
    while (TRUE) {
        if (length < (p - text)) {
            return (NULL);
        }
        if (l == line) {
            break;
        }
        if (*p == '\n') {
            l++;
        }
        p++;
    }
    
    return (p);
}

#ifdef NOT_TESTED_YET

/* 
 * char *getHeadOfLine(char *text, int length, int line)
 * {
 *     char    *p;
 *     int     l;
 *     
 *     l = 1;
 *     while (*text != '\0') {
 *         if (l == line) {
 *             break;
 *         } else if (*text == '\n') {
 *             l++;
 *         }
 *         text++;
 *     }
 *     return ((*text == '\0') ? NULL : text);
 * }
 */

#endif /* NOT_TESTED_YET */
